from nostalgia.ndf import NDF
from nostalgia.times import datetime_from_timestamp
import urllib.parse


def get_thumbnail(url):
    img = urllib.parse.parse_qs(urllib.parse.urlparse(url).query)
    return "http://i3.ytimg.com/vi/{}/maxresdefault.jpg".format(img["v"][0]) if "v" in img else None


class VideosWatched(NDF):
    vendor = "web"

    @classmethod
    def handle_dataframe_per_file(cls, data, fname):
        data["playingSince"] = [datetime_from_timestamp(x, "utc") for x in data["playingSince"]]
        data["playingUntil"] = [datetime_from_timestamp(x, "utc") for x in data["playingUntil"]]
        return data

    @classmethod
    def load(cls, nrows=None):
        data = cls.load_data_file_modified_time(
            "~/nostalgia_data/videos_watched.jsonl", nrows=nrows
        )
        data["image"] = [get_thumbnail(x) for x in data["loc"]]

        return cls(data.reset_index())



from nostalgia.times import datetime_from_format
from datetime import timedelta
from nostalgia.sources.google import Google


def custom_parse(x):
    if not isinstance(x, str):
        return x
    try:
        return datetime_from_format(x, "%Y-%m-%dT%H:%M:%SZ", in_utc=True)
    except:
        return datetime_from_format(x, "%Y-%m-%dT%H:%M:%S.%fZ", in_utc=True)


class AppUsage(Google):
    @classmethod
    def handle_dataframe_per_file(cls, data, file_path):
        data["time"] = [custom_parse(x) for x in data["time"]]
        data = data.rename(columns={"header": "name"})
        return data[["name", "time"]]

    @classmethod
    def load(cls, nrows=None, from_cache=True, **kwargs):
        file_path = "~/nostalgia_data/input/google/Takeout/My Activity/Android/My Activity.json"
        data = cls.load_data_file_modified_time(file_path, nrows=nrows, from_cache=from_cache)
        return cls(data)



import os
from nostalgia.ndf import NDF


class Screenshots(NDF):
    @classmethod
    def ingest(cls, file_dir, nrows=None):
        """ Process a picture folder. Takes ~1s per image to OCR. """
        globs = [os.path.join(file_dir, "*.png"), os.path.join(file_dir, "*.jpg")]
        pic_data = cls.load_image_texts(globs, nrows=nrows)
        cls.save_df(pic_data)



from nostalgia.times import tz
import pandas as pd
import just
from nostalgia.interfaces.chat import ChatInterface
from nostalgia.data_loading import read_array_of_dict_from_json
from nostalgia.sources.facebook import Facebook

# # # # #


class FacebookChat(Facebook, ChatInterface):
    """# Facebook Chat
       Facebook allows you to export all your data. This source is about the chat between two people using Facebook Messenger.

       ### Obtaining the data
       Go to https://www.facebook.com/settings?tab=your_facebook_information and click "Download your information".

       Make sure to select at least **Messages**, and choose **JSON** format above.

       Afterwards, unpack the ZIP-archive and provide the root folder as "file_path".
       For "user", fill in a subfolder name (e.g. johnsmith_nx44ludqrh)."""

    me = ""
    sender_column = "sender_name"

    @classmethod
    def load(cls, nrows=None):
        file_path = "~/nostalgia_data/input/facebook"
        chat_paths = just.glob(f"{file_path}/messages/inbox/*/message_1.json")
        face = pd.concat(
            [read_array_of_dict_from_json(chat_file, "messages", nrows) for chat_file in chat_paths]
        )
        face = face.reset_index(drop=True).sort_values("timestamp_ms")
        face["time"] = pd.to_datetime(face["timestamp_ms"], unit='ms', utc=True).dt.tz_convert(tz)
        face.drop("timestamp_ms", axis=1, inplace=True)
        face.loc[
            (face["type"] != "Generic") | face["content"].isnull(), "content"
        ] = "<INTERACTIVE>"
        face["path"] = ""
        if "photos" in face:
            not_null = face["photos"].notnull()
            face.loc[not_null, "path"] = [file_path + x[0]["uri"] for x in face[not_null]["photos"]]
        # if "photos" in face and isinstance(face["photos"]):
        #     face["path"] = [x.get("uri") if x else x for x in face["photos"]]
        return cls(face)



import just
import portray

NDF_PATH = "nostalgia/ndf.py"


def replace_ndf_class():
    ndf_replace = "class NDF:"
    original = just.read(NDF_PATH)
    ndf = [x for x in original.split("\n") if x.startswith("class NDF")][0]
    just.write(original.replace(ndf, ndf_replace), NDF_PATH)
    return original


# remove base pandas from documentation
original_ndf = replace_ndf_class()
try:
    portray.on_github_pages()
except KeyboardInterrupt:
    print("Exiting")
finally:
    # put pandas class back in
    just.write(original_ndf, NDF_PATH)



import os
import sys
import re
import just
import pandas as pd
import numpy as np


from tok import word_tokenize

if sys.argv[-1].endswith("ipython"):
    pd.set_option("display.precision", 10)
    pd.set_option("precision", 10)
    pd.set_option('display.max_columns', 500)
    pd.set_option('display.max_colwidth', -1)

has_alpha = re.compile("[a-zA-Z]")
num = re.compile("[0-9]")


class MockCredentials:
    def __init__(self, credential_dict):
        for k, v in credential_dict.items():
            setattr(self, k, v)


def get_tokens(sentence, lower=True):
    if not isinstance(sentence, str):
        return []
    if sentence is None:
        return []
    return [
        x
        for x in word_tokenize(sentence, to_lower=lower)
        if len(x) > 2 and has_alpha.search(x) and not num.search(x)
    ]


def get_token_set(iterable_of_sentences, lower=True):
    words = set()
    for x in iterable_of_sentences:
        words.update(get_tokens(x, lower))
    return words


def format_latlng(latlng):
    if isinstance(latlng, str):
        latlng = latlng.replace(",", " ").split()
    return "{:0.7f}, {:0.7f}".format(*map(float, latlng))


def view(path):
    from requests_viewer import view_html

    view_html(just.read(path))


def haversine(lat1, lon1, lat2, lon2):
    """
    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees)
    currently in meters
    """

    # Convert decimal degrees to Radians:
    lon1 = np.radians(lon1)
    lat1 = np.radians(lat1)
    lon2 = np.radians(lon2)
    lat2 = np.radians(lat2)

    # Implementing Haversine Formula:
    dlon = np.subtract(lon2, lon1)
    dlat = np.subtract(lat2, lat1)

    a = np.add(
        np.power(np.sin(np.divide(dlat, 2)), 2),
        np.multiply(
            np.cos(lat1), np.multiply(np.cos(lat2), np.power(np.sin(np.divide(dlon, 2)), 2))
        ),
    )
    c = np.multiply(2, np.arcsin(np.sqrt(a)))
    # km
    r = 6371
    # meters
    r *= 1000

    return c * r


seps = [".", ","]


def parse_price(x):
    if x is None:
        return None
    if isinstance(x, list):
        x = "".join(x)
    if isinstance(x, (int, float)):
        if x != x:
            return None
        return x
    x = x.replace(" ", "")
    try:
        return float(x)
    except ValueError:
        pass
    result = np.nan
    for decimal_sep in seps:
        for thousand_sep in seps:
            if decimal_sep == thousand_sep:
                continue
            regex = "^[+-]?"
            regex += "[0-9]{1,3}"
            regex += "(?:[" + thousand_sep + "]?[0-9]{3})*"
            regex += "([" + decimal_sep + "][0-9]{1,2})?"
            regex += "$"
            if re.match(regex, x):
                try:
                    result = float(x.replace(thousand_sep, "").replace(decimal_sep, "."))
                except ValueError:
                    continue
                return result
    return result


def normalize_name(name):
    return re.sub("([A-Z])", r"_\1", name).lstrip("_").lower()


def clean_df(df, dc):
    for k, v in dc.items():
        df[k] = [v(x) for x in df[k]]
    return df


def load_entry():
    import sys

    just.write("", "~/nostalgia_data/__init__.py")

    sys.path.append(os.path.expanduser("~/nostalgia_data/"))
    import nostalgia_entry

    return nostalgia_entry



import re
import just
import pandas
from nostalgia.times import datetime_from_timestamp
from nostalgia.sources.facebook import Facebook
from nostalgia.interfaces.post import PostInterface


class FacebookPosts(Facebook, PostInterface):
    @classmethod
    def handle_json(cls, data):
        posts = []
        for post in data:
            if "data" not in post or not isinstance(post["data"], list):
                continue
            location = "self"
            title = post.get("title", "")
            location_res = re.findall(
                "(?:on|to) ([^']+)'s? [tT]imeline|posted in ([^.]+)|was with ([^.]+)[.]$", title
            )
            if location_res:
                location = [x for x in location_res[0] if x][0]
            for x in post["data"]:
                if "post" not in x:
                    continue
                row = {
                    "location": location,
                    "title": x["post"],
                    "time": datetime_from_timestamp(post['timestamp']),
                }
                posts.append(row)
        return posts

    @classmethod
    def load(cls, nrows=None):
        data = cls.load_json_file_modified_time(
            "~/nostalgia_data/input/facebook/posts/your_posts_1.json"
        )
        return cls(data)



import re
import just
import json
import pandas as pd
import getpass

from nostalgia.ndf import NDF
from nostalgia.times import tz

login_url = "https://s.sleepcycle.com/site/login"
export_url = 'https://s.sleepcycle.com/export/original'


class SleepCycle(NDF):
    @classmethod
    def ingest(cls, credentials, **kwargs):
        html = just.get(login_url)
        token = re.findall('name="csrftoken" value="([^"]+)', html)[0]

        data = {
            "username": credentials.username,
            "csrftoken": token,
            "password": credentials.password,
        }
        _ = just.post(login_url, data=data)

        str_data = just.get('https://s.sleepcycle.com/export/original')

        start = str_data.index("data_json.txt") + len("data_json.txt")
        end = str_data[start:].index("}]PK") + 2

        data = json.loads(str_data[start:][:end])

        xs = []
        ys = []
        nums = []
        for num, x in enumerate(data):
            res = []
            times = []
            if len(x["events"]) < 15:
                continue
            for y in x["events"]:
                res.append(y[-1])
                times.append(y[0])
                nums.append(num)
            xs.extend(
                [
                    pd.Timestamp(x["start"], tz=tz) + pd.Timedelta(seconds=int(y))
                    for y in pd.Series(times).rolling(15).mean().fillna(method="bfill")
                ]
            )
            ys.extend(pd.Series(res).rolling(15).mean().fillna(method="bfill"))
        df = pd.DataFrame({"time": xs, "score": ys, "num": nums}).drop_duplicates()
        df["score"] = df["score"].clip(0, 0.025)

        cls.save_df(df)

        return cls(df)



import re
import just
import numpy as np
import pandas as pd
from nostalgia.ndf import NDF
import lxml.html
import diskcache
from auto_extract import parse_article
from nostalgia.nlp import nlp
from nostalgia.cache import get_cache
from nostalgia.data_loading import read_array_of_dict_from_json
from nostalgia.times import datetime_from_timestamp

CACHE = get_cache("chrome_history")


# def destroy_tree(tree):
#     node_tracker = {tree: [0, None]}

#     for node in tree.iterdescendants():
#         parent = node.getparent()
#         node_tracker[node] = [node_tracker[parent][0] + 1, parent]

#     node_tracker = sorted(
#         [(depth, parent, child) for child, (depth, parent) in node_tracker.items()],
#         key=lambda x: x[0],
#         reverse=True,
#     )

#     for _, parent, child in node_tracker:
#         if parent is None:
#             break
#         parent.remove(child)

#     del tree


def get_title(x):
    if not x.get("domain"):
        return ""
    if x["path"] in CACHE:
        return CACHE[x["path"]]
    # tree = lxml.html.fromstring(just.read(x["path"]))
    # title = tree.xpath("/html/head/title/text()") or tree.xpath("//title/text()") or [""]
    # destroy_tree(tree)
    # title = title[0]
    match = re.search("<title>([^<]+)</title", just.read(x["path"]), re.MULTILINE)
    title = match.groups()[0].strip() if match is not None else ""
    CACHE[x["path"]] = title
    return title


class WebHistory(NDF):
    keywords = ["website", "page", "site", "web history", "page history", "visit"]
    nlp_columns = ["domain", "domain_and_suffix", "title"]
    # selected_columns = ["time", "name", "price", "url"]

    @nlp("filter", "watch", "see", "view")
    def shows(self):
        return self[np.array(self.title.str.extract("Watch (.+) Full Movie").notna())]

    @nlp("filter", "show", "series")
    def series(self):
        shows = self.shows()
        return shows[shows.title.str.contains("Season")]

    @nlp("filter", "movies", "films")
    def movies(self):
        shows = self.shows()
        return shows[~shows.title.str.contains("Season")]

    @classmethod
    def object_to_row(cls, x):
        import tldextract

        if x["url"]:
            extract = tldextract.extract(x["url"])
            x["domain"] = extract.domain
            x["domain_and_suffix"] = extract.domain + "." + extract.suffix
        x["url"] = x["url"] or "_".join(x["path"].split("_")[1:])
        x["time"] = datetime_from_timestamp(x["time"])
        x["title"] = get_title(x)
        return x

    @classmethod
    def load(cls, file_path="~/nostalgia_data/meta.jsonl", nrows=None, **kwargs):
        web_history = cls.load_object_per_newline(file_path, nrows)
        return cls(web_history)



from nostalgia.times import datetime_from_timestamp
from nostalgia.interfaces.post import PostInterface


class RedditPosts(PostInterface):
    vendor = "reddit"

    @classmethod
    def ingest(cls, author):
        from psaw import PushshiftAPI

        api = PushshiftAPI()

        posts = [
            {
                "title": x.title,
                "time": datetime_from_timestamp(x.created_utc),
                "url": x.full_link,
                "text": x.selftext,
                "author": author,
            }
            for x in api.search_submissions(author=author)
        ]

        cls.save_df(posts)



import os
import re
import pandas as pd
from datetime import datetime
from nostalgia.times import tz

from nostalgia.ndf import NDF

# Add this to your emacs startup init:
# (require 'f)
# (defun log-find-visits ()
#   (when (and buffer-file-name (not (eq last-command "xah-close-current-buffer")))
#       (f-append-text (concat (int-to-string (float-time)) "," buffer-file-name "\n") 'utf-8 "~/nostalgia_data/input/log-emacs-find-visits.txt")))


class FileVisits(NDF):
    @classmethod
    def load(cls, file_name="~/nostalgia_data/input/log-emacs-find-visits.txt", nrows=None):
        fname = os.path.expanduser(file_name)
        with open(fname) as f:
            results = []
            files = []
            ds = []
            num = 0
            for line in f.read().split("\n"):
                if not line:
                    continue
                y, z = line.split(",", 1)
                if not y:
                    continue
                files.append(z)
                d = datetime.fromtimestamp(float(y), tz)
                ds.append(d)
                for key in ["egoroot/", "/ssh:", "gits/", "site-packages/", "Drive/", "Dropbox/"]:
                    if key in z:
                        z = re.split("[:/]", z.split(key)[1])[0]
                results.append(z)
                if num == nrows:
                    break
                num += 1
        data = pd.DataFrame({"file": files, "name": results, "time": ds})
        return cls(data)



import just
import pandas as pd
from datetime import datetime
from nostalgia.times import tz, parse
from nostalgia.data_loading import read_array_of_dict_from_json
from nostalgia.sources.google import Google


def try_parse(x):
    try:
        if x.endswith("PST"):
            x = x.replace("PST", "-0800 (PST)")
        elif x.endswith("PDT"):
            x = x.replace("PDT", "-0700 (PDT)")
        d = parse(x)
        if d.tzinfo is None:
            d = d.replace(tzinfo=tz)
        return d
    except:
        return datetime(1970, 1, 1, 0, 0, 0, tzinfo=tz)


class Gmail(Google):
    me = []

    @classmethod
    def handle_dataframe_per_file(cls, data, fname):
        data["subject"] = data["subject"].astype(str)
        data["to"] = data["to"].astype(str)
        data["sender"] = data["from"].str.extract("<([^>]+)>")
        data.loc[data["sender"].isnull(), "sender"] = data[data["sender"].isnull()][
            "from"
        ].str.strip('"')
        data["sent"] = data.sender.str.contains("|".join(cls.me), na=False)
        data["receiver"] = data["to"].str.extract("<([^>]+)>")
        data.loc[data["receiver"].isnull(), "receiver"] = data.loc[data["receiver"].isnull(), "to"]
        data["received"] = data.receiver.str.contains("|".join(cls.me), na=False)
        data["timestamp"] = pd.to_datetime([try_parse(x) for x in data.date], utc=True).tz_convert(
            tz
        )
        data.drop("date", axis=1, inplace=True)
        return data

    # refactor to use "google_takeout" as target
    @classmethod
    def load(cls, fname, nrows=None, from_cache=True):
        emails = cls.load_data_file_modified_time(fname, nrows=nrows, from_cache=from_cache)
        return cls(emails)

    def sent_by(self, name=None, email=None, case=False):
        if name is not None and email is not None:
            a = self.sender.str.contains(name, case=case, na=False)
            b = self.sender.str.contains(email, case=case, na=False)
            res = self[a | b]
        elif name is not None:
            res = self[self.sender.str.contains(name, case=case, na=False)]
        elif email is not None:
            res = self[self.sender.str.contains(email, case=case, na=False)]
        return self.__class__(res)

    def received_by(self, name=None, email=None, case=False):
        if name is not None and email is not None:
            a = self.receiver.str.contains(name, case=case, na=False)
            b = self.receiver.str.contains(email, case=case, na=False)
            res = self[a | b]
        elif name is not None:
            res = self[self.receiver.str.contains(name, case=case, na=False)]
        elif email is not None:
            res = self[self.receiver.str.contains(email, case=case, na=False)]
        return self.__class__(res)



import pandas as pd
from nostalgia.sources.google import Google
from nostalgia.times import datetime_from_format
from nostalgia.times import tz


class PageVisit(Google):
    @classmethod
    def handle_dataframe_per_file(cls, data, file_path):
        data["time"] = pd.to_datetime(data.time_usec, utc=True, unit="us").dt.tz_convert(tz)
        del data["time_usec"]
        return data

    # refactor to use "google_takeout" as target
    @classmethod
    def load(cls, nrows=None, from_cache=True, **kwargs):
        file_path = "~/nostalgia_data/input/google/Takeout/Chrome/BrowserHistory.json"
        page_visit = cls.load_data_file_modified_time(
            file_path, "Browser History", nrows=nrows, from_cache=from_cache
        )
        return cls(page_visit)



import os
import just
import pandas as pd

from nostalgia.utils import normalize_name

from nostalgia.file_caching import save_df, load_df
from nostalgia.file_caching import get_newline_count, save_newline_count
from nostalgia.file_caching import get_processed_files, save_processed_files
from nostalgia.file_caching import get_last_latest_file, save_last_latest_file
from nostalgia.file_caching import get_last_mod_time, save_last_mod_time


def read_array_of_dict_from_json(fname, key_name=None, nrows=None):
    """
    This is an iterative way to read a json file without having to construct Python elements for everything.
    It can be a lot faster.

    Example data:
    {"participants": {"name": "a", "name": "b", "messages": [{"sender": "a", "time": 123}, {"sender": "b", "time": 124}]}}

    Function call:
    read_array_of_dict_from_json(fname, "messages", nrows=1)

    Returns:
    pd.DataFrame([{"sender": "a", "time": 123}])
    """
    if fname.endswith(".jsonl"):
        if not key_name:
            return pd.read_json(fname, lines=True)
        else:
            return pd.DataFrame([x[key_name] for x in just.read(fname)])

    if nrows is None:
        if not key_name:
            return pd.read_json(fname, lines=fname.endswith(".jsonl"))
        else:
            return pd.DataFrame(just.read(fname)[key_name])

    import ijson

    with open(just.make_path(fname)) as f:
        parser = ijson.parse(f)
        capture = False
        rows = []
        row = {}
        map_key = ""
        num = 0
        for prefix, event, value in parser:
            if num > nrows:
                break
            if prefix == key_name and event == "start_array":
                capture = True
            if not capture:
                continue
            if event == "start_map":
                continue
            elif event == "map_key":
                map_key = value
            elif event == "end_map":
                rows.append(row)
                row = {}
                num += 1
            elif map_key:
                row[map_key] = value
    return pd.DataFrame(rows)


class Loader:
    @classmethod
    def load_data_file_modified_time(
        cls, fname, key_name="", nrows=None, from_cache=True, **kwargs
    ):
        """
        It will load from cache if filename is not changed since last run (and there is a cache).
        If it has changed, it will reprocess and save it in cache (including the modified_time).
        Handles csv, mbox and json currently.
        key_name is only for json.
        nrows is for enabling quickly loading a sample.
        from_cache=False allows ignoring the cache and reprocessing the file.

        Loading the csv, json or mbox file will yield you a DF
        IMPORTANT: assumes you implement `handle_dataframe_per_file`
        This is the post-processing required after the file is loaded, for e.g. converting time
        dropping and adding columns.
        """
        name = fname + "_" + normalize_name(cls.__name__)
        modified_time = os.path.getmtime(os.path.expanduser(fname))
        last_modified = get_last_mod_time(name)
        if modified_time != last_modified or not from_cache:
            if fname.endswith(".csv"):
                data = pd.read_csv(fname, error_bad_lines=False, nrows=nrows, **kwargs)
            elif fname.endswith(".mbox"):
                import mailbox

                m = mailbox.mbox(fname)
                data = pd.DataFrame(
                    [{l: x[l] for l in ["from", "to", "date", "subject"]} for x in m]
                )
            else:
                data = read_array_of_dict_from_json(fname, key_name, nrows, **kwargs)
            data = cls.handle_dataframe_per_file(data, fname)
            if nrows is None:
                save_df(data, name)
                save_last_mod_time(modified_time, name)
        else:
            data = load_df(name, nrows)
        if nrows is not None:
            data = data.iloc[-nrows:]
        return data

    @classmethod
    def load_json_file_modified_time(cls, fname, nrows=None, from_cache=True, **kwargs):
        name = fname + "_" + normalize_name(cls.__name__)
        modified_time = os.path.getmtime(os.path.expanduser(fname))
        last_modified = get_last_mod_time(name)
        if modified_time != last_modified or not from_cache:
            data = just.read(fname)
            data = cls.handle_json(data, **kwargs)
            data = pd.DataFrame(data)
            if nrows is None:
                save_df(data, name)
                save_last_mod_time(modified_time, name)
        else:
            data = load_df(name)
        if nrows is not None:
            data = data.iloc[-nrows:]
        return data

    @classmethod
    def load_image_texts(cls, glob_pattern_s, nrows=None):
        import pytesseract
        from PIL import Image

        if isinstance(glob_pattern_s, list):
            fnames = set()
            for glob_pattern in glob_pattern_s:
                fnames.update(set(just.glob(glob_pattern)))
            glob_pattern = "_".join(glob_pattern_s)
        else:
            fnames = set(just.glob(glob_pattern))
        name = glob_pattern + "_" + normalize_name(cls.__name__)
        processed_files = get_processed_files(name)
        to_process = fnames.difference(processed_files)
        objects = []

        cache = get_cache("tesseract")

        if nrows is not None:
            if not to_process:
                return load_df(name).iloc[-nrows:]
            else:
                to_process = list(to_process)[-nrows:]
        if to_process:
            for fname in to_process:
                if fname in cache:
                    text = cache[fname]
                else:
                    try:
                        text = pytesseract.image_to_string(Image.open(just.make_path(fname)))
                    except OSError as e:
                        print("ERR", fname, e)
                        continue
                    cache[fname] = text
                time = datetime_from_timestamp(os.path.getmtime(fname), "utc")
                data = {"text": text, "path": fname, "title": fname.split("/")[-1], "time": time}
                objects.append(data)
            data = pd.DataFrame(objects)
            if processed_files and nrows is None:
                data = pd.concat((data, load_df(name)))
            for x in ["time", "start", "end"]:
                if x in data:
                    data = data.sort_values(x)
                    break
            if nrows is None:
                save_df(data, name)
                save_processed_files(fnames | processed_files, name)
        else:
            data = load_df(name)
        if nrows is not None:
            data = data.iloc[-nrows:]
        return data

    @classmethod
    def load_dataframe_per_json_file(cls, glob_pattern, key="", nrows=None):
        fnames = set(just.glob(glob_pattern))
        name = glob_pattern + "_" + normalize_name(cls.__name__)
        processed_files = get_processed_files(name)
        to_process = fnames.difference(processed_files)
        objects = []
        if nrows is not None:
            if not to_process:
                to_process = list(processed_files)[-nrows:]
            else:
                to_process = list(to_process)[-nrows:]
        if to_process:
            print("processing {} files".format(len(to_process)))
            for fname in to_process:
                data = read_array_of_dict_from_json(fname, key, nrows)
                data = cls.handle_dataframe_per_file(data, fname)
                if data is None:
                    continue
                objects.append(data)
            data = pd.concat(objects)
            if processed_files and nrows is None:
                data = pd.concat((data, load_df(name)))
            for x in ["time", "start", "end"]:
                if x in data:
                    data = data.sort_values(x)
                    break
            if nrows is None:
                save_df(data, name)
                save_processed_files(fnames | processed_files, name)
        else:
            data = load_df(name)
        if nrows is not None:
            data = data.iloc[-nrows:]
        return data

    @classmethod
    def load_object_per_newline(cls, fname, nrows=None):
        """
        Iterates over a file containing an object per line (e.g. .jsonl or .txt).
        Will only handle new lines not seen earlier; it detects this by storing the number-of-objects seen.
        You should implement `object_to_row(cls, row)` on your class that returns a dictionary.
        """
        data = []
        name = fname + "_" + normalize_name(cls.__name__)
        newline_count = get_newline_count(name)
        for i, x in enumerate(just.iread(fname)):
            if nrows is None:
                if i < newline_count:
                    continue
            row = cls.object_to_row(x)
            if row is None:
                continue
            data.append(row)
            # breaking at approx 5 rows
            if nrows is not None and i > nrows:
                break
        if data:
            data = pd.DataFrame(data)
            if newline_count and nrows is None:
                data = pd.concat((data, load_df(name)))
            if nrows is None:
                data = save_df(data, name)
                n = i + 1
                save_newline_count(n, name)
        else:
            data = load_df(name)
        if nrows is not None:
            data = data.iloc[-nrows:]
        return data

    @classmethod
    def latest_file_is_historic(cls, glob, key_name="", nrows=None, from_cache=True):
        """
        Glob is for using a wildcard pattern, and the last created file will be loaded.
        See `load_data_file_modified_time` for further reference.
        Returns a pd.DataFrame
        """
        recent = max([x for x in just.glob(glob) if "(" not in x], key=os.path.getctime)
        return cls.load_data_file_modified_time(recent, key_name, nrows, from_cache)



# <Placemark>
#   <name>Home (Laan van Henegouwen 16)</name>
#   <address>3703 TD Zeist</address>
#   <ExtendedData>
#     <Data name="Email">
#       <value>kootenpv@gmail.com</value>
#     </Data>
#     <Data name="Category">
#       <value/>
#     </Data>
#     <Data name="Distance">
#       <value>0</value>
#     </Data>
#   </ExtendedData>
#   <description>  from 2019-01-04T16:41:11.016Z to 2019-01-05T16:29:16.926Z. Distance 0m </description>
#   <Point>
#     <coordinates>5.2261821,52.080966499999995,0</coordinates>
#   </Point>
#   <TimeSpan>
#     <begin>2019-01-04T16:41:11.016Z</begin>
#     <end>2019-01-05T16:29:16.926Z</end>
#   </TimeSpan>
# </Placemark>

import pandas as pd
import lxml.etree
from dateutil.parser import parse as date_parse
import just
from nostalgia.times import yesterday
from nostalgia.utils import format_latlng

N = {"klm": "http://www.opengis.net/kml/2.2"}

# cats = set()
days = []
for fname in sorted(just.glob("/home/pascal/Downloads/ghistory/history-*.kml")):
    # for fname in sorted(["/home/pascal/Downloads/ghistory/history-2018-07-25.kml"]):
    tree = lxml.etree.parse(fname)
    for placemark in tree.xpath("//klm:Placemark", namespaces=N):
        name = placemark.xpath("./klm:name/text()", namespaces=N)[0]
        address = placemark.xpath("./klm:address/text()", namespaces=N)
        address = address[0] if address else None
        start = date_parse(placemark.xpath("./klm:TimeSpan/klm:begin/text()", namespaces=N)[0])
        end = date_parse(placemark.xpath("./klm:TimeSpan/klm:end/text()", namespaces=N)[0])
        category = placemark.xpath(
            "./klm:ExtendedData/klm:Data[@name='Category']/klm:value/text()", namespaces=N
        )
        category = category[0] if category else None
        # just a location
        if category is None and address is None:
            continue
        if category is None and " (" in name:
            category = name.split(" (")[0]
        if "Steynlaan" in name:
            category = "Home"
        distance = placemark.xpath(
            "./klm:ExtendedData/klm:Data[@name='Distance']/klm:value/text()", namespaces=N
        )
        distance = distance[0] if distance else None
        # fuckers, lon|lat instead of lat|lon
        coords = placemark.xpath("./klm:Point/klm:coordinates/text()", namespaces=N)
        coords = coords or placemark.xpath("./klm:LineString/klm:coordinates/text()", namespaces=N)
        d = {
            "date": start.date(),
            "name": name,
            "distance": distance,
            "category": category,
            "start": start,
            "end": end,
            "address": address,
            "month": start.month,
            "hour": start.hour,
            "year": start.year,
        }
        loc = {"lat": None, "lon": None, "coordinates": []}
        # here switch lon|lat
        coor = sum(
            [
                [format_latlng((x.split(",")[1], x.split(",")[0])).split(", ") for x in c.split()]
                for c in coords
            ],
            [],
        )
        if coor:
            loc["coordinates"] = coor
            loc["lat"] = coor[0][0]
            loc["lon"] = coor[0][1]
        # for data in placemark.xpath("//klm:Data", namespaces=N):
        #     cats.update(data.xpath("./@name", namespaces=N))
        d.update(loc)
        days.append(d)

df = pd.DataFrame(days)
df = df.reindex(list(d.keys()), axis=1)


yday = yesterday()
tmpl = "/home/pascal/Downloads/timeline_data-{}-{:02d}-{:02d}.csv"
df.to_csv(tmpl.format(yday.year, yday.month, yday.day))



from datetime import datetime
import os
import requests
import diskcache
import dotenv
import hashlib
import pandas as pd
from pytz import timezone
import just
from nostalgia.times import tz
from nostalgia.utils import format_latlng
from nostalgia.cache import get_cache
from nostalgia.interfaces.places import Places


def api_call(url, params):
    key = get_hash(url, params)
    status = CACHE.get(key, {}).get("status")
    if status == "ZERO_RESULTS":
        return None
    if status == "OK":
        return CACHE[key]
    jdata = s.get(url, params=params).json()
    jdata["meta"] = {
        "time": str(datetime.utcnow()),
        "url": url,
        "params": {k: v for k, v in params.items() if k != "key"},
    }
    if jdata.get("status") == "ZERO_RESULTS":
        CACHE[key] = jdata
        return None
    if jdata.get("status") != "OK":
        print("status", url, params, key, jdata.get("status"))
        raise ValueError("not ok")
    CACHE[key] = jdata
    return jdata


def get_nearby(latlng, name, excluded_transport_names):
    latlng = format_latlng(latlng)
    if latlng == "nan, nan":
        return None
    params = {"location": latlng, "radius": 100, "key": KEY}
    if name not in excluded_transport_names:
        params["query"] = name
    return api_call(NEARBY_URL, params)


def get_nearby_results(latlng, name, excluded_transport_names):
    near = get_nearby(latlng, name, excluded_transport_names)
    if near is None:
        return None
    res = [x for x in near["results"] if name in excluded_transport_names or x["name"] == name]
    for x in res:
        if "opening_hours" in x:
            return x
    if res:
        return res[0]
    if not near["results"]:
        return None
    return near["results"][0]


def get_details(place_id):
    params = {'placeid': place_id, 'sensor': 'false', 'key': KEY}
    return api_call(DETAILS_URL, params).get("result", {})


def get_(details, *types):
    for tp in types:
        for addr in details.get('address_components', []):
            if tp in addr["types"]:
                return addr["long_name"]


def get_hash(url, params):
    data = {k: v for k, v in params.items() if k != "key"}
    data["url"] = url
    data = sorted(data.items())
    return hashlib.sha256(str(data).encode("utf8")).hexdigest()


GEOCODE_URL = "https://maps.googleapis.com/maps/api/geocode/json"


def get_address(json_response):
    for res in json_response.get("results", []):
        addr = res.get('formatted_address')
        if addr:
            return addr


def geo_get_(json_response, *types):
    for res in json_response.get("results", []):
        for tp in types:
            for addr in res.get('address_components', []):
                if tp in addr["types"]:
                    return addr["long_name"]


def geo_get_info(latlng):
    latlng = format_latlng(latlng)
    if latlng == "nan, nan":
        return None
    params = {"method": "reverse", "latlng": latlng, "key": KEY}
    json_response = api_call(GEOCODE_URL, params=params)
    city = geo_get_(json_response, "locality", "postal_town")
    # place_id = json_response.get("results", [{}])[0].get("place_id")
    country = geo_get_(json_response, "country")
    address = get_address(json_response)
    return {"city": city, "country": country, "formatted_address": address}


dotenv.load_dotenv("google/.env")
dotenv.load_dotenv(".env")

PYTHON_ENV = os.environ.get("PYTHON_ENV", "dev")
if PYTHON_ENV != "prod":
    KEY = None
else:
    KEY = os.environ.get("GOOGLE_API_KEY", None)

CACHE = get_cache("google_timeline")

DETAILS_URL = "https://maps.googleapis.com/maps/api/place/details/json"
NEARBY_URL = 'https://maps.googleapis.com/maps/api/place/nearbysearch/json'

s = requests.Session()
# {'Boating', 'Cycling', 'Driving', 'Flying', 'In transit', 'Moving', 'On a bus', 'On a ferry', 'On a train', 'On a tram', 'On the subway', 'Running', 'Walking'}


def get_results(latlng, name, excluded_transport_names):
    if name in excluded_transport_names:
        return geo_get_info(latlng)
    near_result = get_nearby_results(latlng, name, excluded_transport_names)
    if near_result is None:
        return None
    details = get_details(near_result["place_id"])
    if details is None:
        return None
    return {
        "city": get_(details, "locality", "postal_town"),
        "country": get_(details, "country"),
        "rating": details,
        "details_name": details.get("name"),
        "formatted_address": details.get("formatted_address"),
        'international_phone_number': details.get('international_phone_number'),
        'opening_hours': details.get('opening_hours'),
        'user_ratings_total': details.get('user_ratings_total'),
        'rating': details.get('rating'),
        'website': details.get('website'),
    }


def process(df, excluded_transport_names, home_regex, work_regex, hometown_regex):
    df["start"] = pd.to_datetime(df.start)
    df["end"] = pd.to_datetime(df.end)
    df["around_noon"] = (df.start.dt.hour < 12) & (df.end.dt.hour > 12)
    df["week"] = df.start.dt.weekday < 5
    df["city"] = df.city.fillna("")
    df["transporting"] = df.category.isin(excluded_transport_names)

    if work_regex:
        df.loc[df.name.str.contains(work_regex, regex=True, na=False), "category"] = "Work"
    if home_regex:
        df.loc[df.name.str.contains(home_regex, regex=True, na=False), "category"] = "Home"
    # if hometown_regex:
    #     df.loc[
    #         df.name.str.contains(hometown_regex, regex=True, na=False), "category"
    #     ] = "Hometown"

    df = df.sort_values("start").reset_index(drop=True)
    if "Unnamed: 0" in df.columns:
        del df["Unnamed: 0"]
    df["start"] = df["start"].dt.tz_convert(tz)
    df["end"] = df["end"].dt.tz_convert(tz)
    df.index = pd.IntervalIndex.from_arrays(df['start'], df['end'])
    return df


class GooglePlaces(Places):
    nlp_columns = ["name", "city", "country", "category", "website"]
    selected_columns = ["date", "name", "city", "_office_hours"]

    @classmethod
    def load(cls, file_glob="~/Downloads/timeline_data-*", nrows=None):
        df = pd.read_csv(max(just.glob(file_glob)), nrows=nrows)

        unique_locs = set([((y, z), x) for x, y, z in zip(df.name, df.lat, df.lon) if y != "nan"])

        excluded_transport_names = set(df[df.name == df.category].name)

        details_data = []
        for (lat, lon), name in unique_locs:
            d = get_results((lat, lon), name, excluded_transport_names)
            if d is None:
                continue
            d["lat"] = lat
            d["lon"] = lon
            d["name"] = name
            details_data.append(d)

        details_data = pd.DataFrame(details_data)

        # all_loc = df.merge(details_data, on=["name", "lat", "lon"], how="outer")
        places = df.merge(details_data, on=["name", "lat", "lon"], how="inner")

        # all_loc = process(all_loc, excluded_transport_names)

        home_regex = "|".join(cls.home)
        work_regex = "|".join(cls.work)
        hometown_regex = "|".join(cls.hometown)
        places = process(places, excluded_transport_names, home_regex, work_regex, hometown_regex)

        return cls(places)


GooglePlaces.work = ["Jibes", "Sleepboot", "De Meerpaal", "Papendorp", "Vektis", "Work "]


# at_work = places.at_work()
#
# places.travel_by_car()\
#       .duration_longer_than(minutes=30)\
#       .at_day(at_work)
#       .at_time("november 2019")\
#       .during_weekdays\

# cat_str = ", ".join(["{}: {}".format(i, x) for i, x in enumerate(places.category)])
# [places.iloc[[int(y) for y in x]] for x in re.findall(", ".join(["([0-9]+): {}".format(x) for x in pats]), cat_str)]



import getpass
from selenium import webdriver
from nostalgia.utils import MockCredentials


def get_driver(
    executable_path="chromedriver",
    binary_location=None,
    headless=False,
    credentials=None,
    login_url=None,
    user_xpath_or_id="username",
    password_xpath_or_id="password",
    **options,
):
    # set options
    chrome_options = webdriver.ChromeOptions()
    if headless:
        chrome_options.add_argument('headless')
    if binary_location is not None:
        setattr(chrome_options, "binary_location", binary_location)
    for k, v in options.items():
        setattr(chrome_options, k, v)
    driver = webdriver.Chrome(executable_path=executable_path, options=chrome_options)

    # login if neccessary
    if isinstance(credentials, dict):
        credentials = MockCredentials(credentials)
    if credentials is not None:
        driver.get(login_url)
        if user_xpath_or_id.startswith("/"):
            user_element = driver.find_elements_by_xpath(user_xpath_or_id)[0]
        else:
            user_element = driver.find_element_by_id(user_xpath_or_id)
        if password_xpath_or_id.startswith("/"):
            password_element = driver.find_elements_by_xpath(password_xpath_or_id)[0]
        else:
            password_element = driver.find_element_by_id(password_xpath_or_id)
        user_element.send_keys(credentials.username)
        password = credentials.password
        password_element.send_keys(password)  # h
        password_element.submit()

    return driver



# ensure ~/nostalgia_data/input exists (e.g. "mkdir -p ~/nostalgia_data/input" on linux)
# goto https://shazam.com/myshazam
# open network tab
# login
# search for url containing "discovery"
# right click and copy as curl and replace limit=20 with limit=2000
# take that curl command and add the following: > ~/nostalgia_data/input/shazam.json and hit return

import pandas as pd
import just
from nostalgia.ndf import NDF
from nostalgia.times import datetime_from_timestamp


class Shazam(NDF):
    @classmethod
    def load(cls, file_path="~/nostalgia_data/input/shazam.json", nrows=None):
        shazam = pd.DataFrame(
            [
                (
                    datetime_from_timestamp(x["timestamp"], x["timezone"]),
                    x["track"]["heading"]["title"],
                    x["track"]["heading"]["subtitle"],
                )
                for x in just.read(file_path)["tags"]
            ],
            columns=["time", "title", "artist"],
        )
        return cls(shazam)



import tqdm
import sys
import just
import zipfile

if __name__ == "__main__":
    for x in tqdm.tqdm(sys.argv[1:]):
        with zipfile.ZipFile(x, "r") as zip_ref:
            zip_ref.extractall()
        just.remove(x)



import re
import just

import urllib.parse

import pandas as pd

from auto_extract import parse_article

from nostalgia.cache import get_cache

from datetime import datetime
from nostalgia.times import tz
from nostalgia.ndf import NDF
from nostalgia.nlp import nlp

CACHE = get_cache("linked_data_videos")


def get_linked_data(x):
    path = x["path"]
    if path in CACHE:
        return CACHE[path]
    try:
        html = just.read(path)
    except EOFError:
        CACHE[path] = None
        return None
    if not html.strip():
        CACHE[path] = None
        return None
    art = parse_article(html, x["url"])
    if "youtube" not in art.domain:
        return None
    title = re.sub(" - YouTube$", "", art.tree.xpath("//title/text()")[0])
    if title == "YouTube":
        CACHE[path] = None
        return None
    if not title:
        return None
    vc = art.tree.xpath("//span[contains(@class, 'view-count')]/text()")
    vc = re.sub("[^0-9]", "", vc[0]) if vc else None
    watch_part = urllib.parse.parse_qs(urllib.parse.urlparse(x["url"]).query)["v"]
    if watch_part:
        image = "http://i3.ytimg.com/vi/{}/maxresdefault.jpg".format(watch_part[0])
    else:
        image = None
    channel = art.tree.xpath("//ytd-video-owner-renderer//a/text()")
    if not channel:
        channel = art.tree.xpath("//ytd-channel-name//a/text()")
    channel = " ".join(channel)
    linked_data = {
        "title": title,
        "type": "video",
        "source": "youtube",
        "image": image,
        "view_count": vc,
        "channel": channel,
    }
    CACHE[path] = linked_data
    return linked_data


class Videos(NDF):
    vendor = "web"

    keywords = ["watch", "viewers", "listen", "video", "music video", "youtube"]
    nlp_columns = ["title", "channel"]
    selected_columns = ["time", "title", "view_count", "url"]

    @classmethod
    def object_to_row(cls, obj):
        url = str(obj["url"])
        if "youtub" not in url or "watch" not in url or "search_query" in url:
            return None
        row = get_linked_data(obj)
        if row is not None:
            row["time"] = datetime.fromtimestamp(float(obj["time"]), tz=tz)
            row["url"] = obj["url"]
            row["path"] = obj["path"]
            row["keywords"] = ""
            return row

    @classmethod
    def load(cls, fname="~/nostalgia_data/meta.jsonl", nrows=None):
        data = cls.load_object_per_newline(fname, nrows)
        return cls(data)

    @nlp("end", "how many views")
    def sum(self):
        return self.view_count.sum()



from pytz import timezone
from pytz.reference import Local
from metadate import parse_date, Units
from datetime import datetime
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta
import dateutil

# tz = timezone('Europe/Amsterdam')
tz = timezone(Local.tzname(datetime.now()))
utc = timezone('UTC')


def now(**kwargs):
    return datetime.now(tz=tz) - relativedelta(**kwargs)


def last_days(days):
    return now(days=days).replace(hour=0, minute=0, second=0, microsecond=0)


def today():
    return last_days(0)


def yesterday():
    return last_days(1)


def last_week():
    return now(days=7).replace(hour=0, minute=0, second=0, microsecond=0)


def last_month():
    return now(months=1).replace(day=1, hour=0, minute=0, second=0, microsecond=0)


def last_year():
    return now(years=1).replace(month=1, day=1, hour=0, minute=0, second=0, microsecond=0)


def week_ago(weeks):
    begin = now(days=7 * weeks).date()
    end = now(days=7 * (weeks - 1)).date()
    return begin, end


def days_ago(days):
    begin = now(days=days).date()
    end = now(days=days - 1).date()
    return lambda: (begin, end)


def months_ago(months):
    begin = now(months=months).date()
    end = now(months=months - 1).date()
    return lambda: (begin, end)


def years_ago(years):
    begin = now(years=years).date()
    end = now(years=years - 1).date()
    return lambda: (begin, end)


def in_month(month):
    year = now().year - 1 if now().month <= month else now().year
    return lambda: (
        datetime(year, month, 1).date(),
        datetime(year, month, 1).date() + relativedelta(month=month + 1),
    )


def in_year(year):
    return lambda: (datetime(year, 1, 1).date(), datetime(year + 1, 1, 1).date())


def try_date(x, min_level=None, max_level=None):
    try:
        # apparently one col is dayfirst, the other isnt
        mps = parse_date(x, multi=True)
        for mp in mps:
            if mp is None:
                continue
            if Units.DAY not in mp.levels:
                continue
            if min_level is not None and mp.min_level > min_level:  # above minute level
                continue
            if max_level is not None and mp.max_level < max_level:
                continue
            date = mp.start_date
            if date.tzinfo is None:
                date = tz.localize(date)
            return date
    except:
        return None


def parse_date_tz(text):
    mp = parse_date(text)
    if mp is not None:
        if mp.start_date.tzinfo is None:
            mp.start_date = tz.localize(mp.start_date)
        if mp.end_date.tzinfo is None:
            mp.end_date = tz.localize(mp.end_date)
        return mp


def datetime_tz(*args):
    return tz.localize(datetime(*args))


def datetime_from_timestamp(x, tzone=tz):
    # would be year 3000
    x = float(x)
    if x > 32503683600:
        x = x // 1000
    if isinstance(tzone, str):
        tzone = timezone(tzone)
    x = datetime.fromtimestamp(x, tz=tzone)
    if tzone != tz:
        x = x.astimezone(tz)
    return x


def datetime_from_format(s, fmt, in_utc=False):
    base = datetime.strptime(s, fmt)
    if in_utc:
        return utc.localize(base).astimezone(tz)
    else:
        return tz.localize(base)



""" File unrelated to the package, except for convenience in deploying """
import re
import sh
import os

commit_count = sh.git("rev-list", ["--all"]).count("\n")

with open("setup.py") as f:
    setup = f.read()

setup = re.sub('MICRO_VERSION = "[0-9]+"', 'MICRO_VERSION = "{}"'.format(commit_count), setup)

major = re.search('MAJOR_VERSION = "([0-9]+)"', setup).groups()[0]
minor = re.search('MINOR_VERSION = "([0-9]+)"', setup).groups()[0]
micro = re.search('MICRO_VERSION = "([0-9]+)"', setup).groups()[0]
version = "{}.{}.{}".format(major, minor, micro)

with open("setup.py", "w") as f:
    f.write(setup)

with open("nostalgia/__init__.py") as f:
    init = f.read()

with open("nostalgia/__init__.py", "w") as f:
    f.write(re.sub('__version__ = "[0-9.]+"', '__version__ = "{}"'.format(version), init))

os.system("rm -rf dist/")
os.system("python setup.py sdist bdist_wheel")
os.system("twine upload dist/*")



import re
import os
import pandas as pd
import just
from nostalgia.times import datetime_from_timestamp, tz
from nostalgia.ndf import NDF
from nostalgia.sources.samsung import Samsung


class SamsungHeartrate(Samsung, NDF):
    @classmethod
    def handle_dataframe_per_file(cls, df, fname):
        if df.empty:
            return None
        df["start"] = [
            datetime_from_timestamp(x) if isinstance(x, int) else tz.localize(x)
            for x in df.start_time
        ]
        df["end"] = [
            datetime_from_timestamp(x) if isinstance(x, int) else tz.localize(x)
            for x in df.end_time
        ]
        del df["start_time"]
        del df["end_time"]
        return df

    @classmethod
    def load(cls, nrows=None, **kwargs):
        file_glob = "~/nostalgia_data/input/samsung/samsunghealth_*/jsons/com.samsung.shealth.tracker.heart_rate"
        file_glob += "/*.com.samsung.health.heart_rate.binning_data.json"
        # TODO FIX
        heartrate = cls.load_dataframe_per_json_file(file_glob, nrows=nrows)
        start = heartrate["start"]
        end = heartrate["end"]
        interval_index = pd.IntervalIndex.from_arrays(start, end)
        heartrate = pd.DataFrame(heartrate)
        heartrate = heartrate.set_index(interval_index).sort_index()
        heartrate["end"] = heartrate.index.right - pd.Timedelta(seconds=1)
        # heartrate = heartrate[heartrate.time != heartrate.end]
        # heartrate = heartrate[heartrate.time <= heartrate.end]
        # if nrows is not None:
        #     heartrate = heartrate.iloc[:nrows]
        heartrate["value"] = heartrate.heart_rate
        del heartrate["heart_rate"]
        return cls(heartrate)



import just
import livereload
import portray

NDF_PATH = "nostalgia/ndf.py"


def ignore(x):
    return "lycheck" in x


def replace_ndf_class():
    ndf_replace = "class NDF:"
    original = just.read(NDF_PATH)
    ndf = [x for x in original.split("\n") if x.startswith("class NDF")][0]
    just.write(original.replace(ndf, ndf_replace), NDF_PATH)
    return original


def render_as_html():
    # remove base pandas from documentation
    original_ndf = replace_ndf_class()
    try:
        portray.as_html(overwrite=True)
    except KeyboardInterrupt:
        print("Exiting")
    # put pandas class back in
    finally:
        just.write(original_ndf, NDF_PATH)


server = livereload.Server()
server.watch("README.md", render_as_html, ignore=ignore)
server.watch("docs/**", render_as_html, ignore=ignore)
server.watch("docs/**/*.md", render_as_html, ignore=ignore)
server.serve(root="site")



import os
import json
from collections import Counter
from datetime import datetime
from urllib.parse import urljoin
import just
import pandas as pd

from nostalgia.utils import parse_price
from nostalgia.times import tz
from nostalgia.nlp import nlp
from nostalgia.ndf import NDF

from auto_extract import parse_article
from nostalgia.sources.web.get_keywords_for_product import get_keywords_for_product

from nostalgia.cache import get_cache

CACHE = get_cache("linked_offers")


def getter(dc, key, default=None):
    res = dc.get(key, default)
    if isinstance(res, list):
        res = res[0]
    elif isinstance(res, dict):
        res = json.dumps(res)
    return res


from natura import Finder

finder = Finder()


def get_linked_data_jd(art):
    data = None
    try:
        jdata = art.jsonld
    except json.JSONDecodeError:
        return None
    for y in jdata:
        if not y:
            continue
        if isinstance(y, list):
            y = y[0]
        if y.get("@type") != "Product":
            continue
        offers = y.get("offers", [])
        if isinstance(offers, dict):
            offers = [offers]
        brand = y.get("brand", {})
        if isinstance(brand, dict):
            brand = brand.get("name")
        for offer in offers:
            currency = offer.get('priceCurrency')
            if currency is None:
                continue
            price = offer.get("price") or offer.get("lowPrice")
            price = parse_price(price)
            if price is None or price == 0:
                continue
            name = getter(y, "name")
            if isinstance(name, list):
                name = "".join(name[:1])
            data = {
                "currency": currency,
                "price": price,
                "name": name,
                "origin": art.url,
                "image": getter(y, "image"),
                "brand": brand,
                "description": getter(y, "description"),
                "source": "jsonld",
            }
            return data


def deeper_price_md(art):
    data = None
    for y in art.microdata:
        tp = str(y.get("type", ""))
        if not tp.endswith("/Product"):
            continue
        props = y.get("properties")
        if props is None:
            continue
        name = y.get("properties", {}).get("name") or " ".join(art.title.split())
        if isinstance(name, list):
            name = name[0]
        offers = y.get("properties", {}).get("offers", {})
        offers = props.get("offers", [])
        if isinstance(offers, dict):
            offers = [offers]
        for offer in offers:
            offer = offer.get("properties", offer)
            price = offer.get("price") or offer.get("lowPrice")
            price = parse_price(price)
            if price is None or price == 0:
                continue
            currency = offer.get("priceCurrency")
            if currency is None:
                continue
            brand = y.get("brand") or props.get("brand") or offer.get("brand")
            if isinstance(brand, dict):
                brand = brand.get("properties", {}).get("name")
            image = y.get("image") or props.get("image") or offer.get("image")
            if isinstance(image, list):
                image = image[0]
            if isinstance(image, dict):
                prop = image.get("properties", image)
                image = [v for k, v in prop.items() if "url" in k.lower()]
                if image:
                    image = image[0]
            description = (
                y.get("description") or props.get("description") or offer.get("description")
            )
            if isinstance(description, list):
                description = description[0]
            elif isinstance(description, dict):
                description = json.dumps(description)
            return {
                "currency": currency,
                "price": price,
                "name": name,
                "origin": art.url,
                "image": urljoin(art.url, image),
                "brand": brand,
                "description": description,
                "source": "md",
            }


def get_linked_data_md(art):
    data = None
    for y in art.microdata:
        props = y.get("properties")
        if props is None:
            continue
        tp = str(y.get("type", ""))
        if tp.endswith("/Offer") or tp.endswith("/AggregateOffer"):
            price = props.get("price") or props.get("lowPrice")
            price = parse_price(price)
            if price is None or price == 0:
                continue
            currency = props.get("priceCurrency")
            if currency is None:
                continue
            return {
                "currency": currency,
                "price": price,
                "name": " ".join(art.title.split()),
                "origin": art.url,
                "image": getter(y, "image"),
                "brand": None,
                "description": y.get("description"),
                "source": "md",
            }

        if tp.endswith("/Product"):
            data = deeper_price_md(art)
            if data is not None:
                return data
            continue
        offers = props.get("offers", [])
        if isinstance(offers, dict):
            offers = [offers]
        brand = props.get("brand", {})
        if isinstance(brand, dict):
            brand = brand.get("properties", {}).get("name")
        for offer in offers:
            offer_properties = offer.get("properties")
            if offer_properties is None:
                continue
            currency = offer_properties.get('priceCurrency')
            if currency is None:
                continue
            price = offer_properties.get("price") or offer_properties.get("lowPrice")
            price = parse_price(price)
            if price is None or price == 0:
                continue
            data = {
                "currency": currency,
                "price": price,
                "name": "".join(props.get("name", "")),
                "origin": art.url,
                "image": getter(props, "image"),
                "brand": brand,
                "description": getter(props, "description"),
                "source": "md",
            }
            return data


def get_linked_amazon(art):
    if "www.amazon.com" not in str(art.url):
        return None
    for x in ["account", "cart"]:
        if x in art.url:
            return None
    paths = [
        "//span[@id='priceblock_ourprice']",
        "//div[@id='buyNewSection']",
        "//form//span[contains(@class, 'offer-price')]",
        "//div[@id='olp-upd-new']//span/a",
        "//div[@id='olp-upd-used']//span/a",
        "//div[@id='olp_feature_div']//span/a",
        "//span[contains(@class, 'header-price')]",
    ]
    price = None
    for path in paths:
        res = art.tree.xpath(path)
        if res:
            price = res[0].text_content()
            break
    if price is not None:
        brand = art.tree.xpath("//a[@id='bylineInfo']")
        brand = brand[0].text_content() if brand else None
        if brand is None:
            br = art.tree.xpath("//div[@id='bylineInfo']")
            if br:
                brand = " ".join(br[0].text_content().split()[1:])
        features = art.tree.xpath("//div[@id='feature-bullets']")
        features = features[0].text_content() if features else None
        data = {
            "currency": "USD",
            "price": max([x.value for x in finder.findall(price)]),
            "name": " ".join(art.title.split()),
            "origin": art.url,
            "image": None,
            "brand": brand,
            "description": features,
            "source": "custom_amazon",
        }
        return data


def get_linked_data(x):
    path = x["path"]
    if path in CACHE:
        return CACHE[path]
    try:
        html = just.read(path)
    except EOFError:
        CACHE[path] = None
        return None
    if not html.strip():
        CACHE[path] = None
        return None
    art = parse_article(html, x["url"])
    linked_data = get_linked_amazon(art)
    if linked_data is None:
        linked_data = get_linked_data_md(art)
        if linked_data is None:
            linked_data = get_linked_data_jd(art)
    CACHE[path] = linked_data
    return linked_data


class Offers(NDF):
    vendor = "web"

    keywords = [
        "product",
        "offer",
        "office",
        "did i see",
        "did i look",
        "did i search",
        "looked at",
        "searched for",
        "did i google",
        "cost",
    ]
    nlp_columns = ["name", "brand", "keywords"]
    selected_columns = ["time", "name", "price", "url"]

    @nlp("end", "how much")
    def sum(self):
        return self.price.sum()

    @classmethod
    def object_to_row(cls, obj):
        row = get_linked_data(obj)
        if row is not None:
            row["time"] = datetime.fromtimestamp(float(obj["time"]), tz=tz)
            row["url"] = obj["url"]
            row["path"] = obj["path"]
            if row.get("name"):
                row["keywords"] = " ".join([x[0] for x in get_keywords_for_product(row["name"])])
            else:
                row["keywords"] = ""
            if "image" in row:
                img = row["image"]
                row["image"] = img if isinstance(img, str) or img is None else img["url"]
            return row

    @classmethod
    def load(cls, file_path="~/nostalgia_data/meta.jsonl", nrows=None, **kwargs):
        offers = cls.load_object_per_newline(file_path, nrows)
        return cls(offers)



import time
import os
import shutil
from selenium import webdriver
from calendar import monthrange
from datetime import datetime
from selenium.common.exceptions import (
    NoSuchElementException,
    ElementNotVisibleException,
    ElementClickInterceptedException,
    ElementNotInteractableException,
)

SELENIUM_EXCEPTIONS = (
    NoSuchElementException,
    ElementNotVisibleException,
    ElementClickInterceptedException,
    ElementNotInteractableException,
)

BASE_PATH = os.path.expanduser("~/Downloads")

# chrome_options = webdriver.ChromeOptions()
# chrome_options.binary_location = "/usr/bin/google-chrome-stable"


def move(fname):
    src = os.path.join(BASE_PATH, fname)
    try:
        shutil.move(src, os.path.join(BASE_PATH, "ghistory", fname))
    except FileNotFoundError:
        print("ERROR: file not found", src)


def exists(year, month, day):
    fname = "history-{}-{:02d}-{:02d}.kml".format(year, month, day)
    return os.path.exists(os.path.join(BASE_PATH, "ghistory", fname))


def get_last_day_of_month(year, num_month):
    return monthrange(year, num_month)[1]


def select_year(year):
    driver.find_element_by_class_name("year-picker").click()
    time.sleep(0.3)
    for ele in driver.find_elements_by_xpath("//div[text() = '{}']/parent::*".format(year)):
        try:
            ele.click()
            break
        except SELENIUM_EXCEPTIONS:
            pass
    time.sleep(1)


def select_month(month):
    driver.find_element_by_class_name("month-picker").click()
    time.sleep(0.3)
    for ele in driver.find_elements_by_xpath("//div[text() = '{}']/parent::*".format(month)):
        try:
            ele.click()
            break
        except SELENIUM_EXCEPTIONS:
            pass
    time.sleep(1)


def click_download():
    driver.find_element_by_xpath("//div[@aria-label=' Settings ']").click()
    time.sleep(1)
    for ele in driver.find_elements_by_xpath("//div[text() = 'Export this day to KML']/parent::*"):
        try:
            ele.click()
            break
        except SELENIUM_EXCEPTIONS:
            pass
    time.sleep(1)


MONTHS = [
    "January",
    "February",
    "March",
    "April",
    "May",
    "June",
    "July",
    "August",
    "September",
    "October",
    "November",
    "December",
]

if __name__ == "__main__":
    driver = webdriver.Chrome()  # , options=chrome_options

    url = "https://www.google.com/maps/timeline?pb=!1m2!1m1!1s2019-02-29"

    driver.get(url)

    now = datetime.now()
    today = (now.year, now.month, now.day)

    # wait for being logged in
    while not driver.find_elements_by_xpath(
        "//span[@class='header-feature-name' and text() = ' Timeline ']"
    ):
        time.sleep(2)

    time.sleep(5)

    try:
        last_clicked_year = None
        last_clicked_month = None
        for year in range(now.year, 2014, -1):
            for num_month, month in enumerate(MONTHS[::-1]):
                num_month = 12 - num_month
                for day in range(get_last_day_of_month(year, num_month), 0, -1):
                    date_tuple = (year, num_month, day)
                    if date_tuple >= today:
                        continue
                    if exists(year, num_month, day):
                        continue
                    print(num_month, date_tuple)
                    # select year if neccessary
                    if last_clicked_year != year:
                        select_year(year)
                        last_clicked_year = year
                    # select month if neccessary
                    if last_clicked_month != month:
                        select_month(month)
                        last_clicked_month = month
                    # handle day
                    driver.find_element_by_class_name("day-picker").click()

                    driver.find_element_by_xpath(
                        "//td[@aria-label='{} {}']".format(day, month[:3])
                    ).click()

                    time.sleep(3)

                    click_download()

                    time.sleep(3)

                    fname = "history-{}-{:02d}-{:02d}.kml".format(year, num_month, day)

                    move(fname)
    finally:
        driver.quit()

    from nostalgia.sources.google.timeline_transform import *
    from nostalgia.sources.google.timeline import *



import pandas as pd
import numpy as np
from nostalgia.ndf import NDF, get_type_from_registry
from nostalgia.nlp import nlp
from nostalgia.times import parse_date_tz
from nostalgia.utils import haversine


class Places(NDF):
    keywords = [
        "did i go",
        "i was in",
        "i was at",
        "was i",
        "what place",
        "place",
        "i was at",
        "visit",
        "go",
        "have i been",
        "stay",
    ]
    home = []
    work = []
    hometown = []

    @nlp("filter", "at home")
    def at_home(self):
        return self.__class__(self[self.category == "Home"])

    @nlp("filter", ["near home", "around home"])
    def near_home(self, distance=1000):
        return self.near_(distance, get_type_from_registry("places").at_home)

    @nlp("filter", "at work")
    def at_work(self):
        return self.__class__(self[self.category == "Work"])

    @nlp("filter", ["near work", "around work"])
    def near_work(self, distance=1000):
        return self.near_(distance, get_type_from_registry("places").at_work)

    @property
    def at_hometown(self):
        raise ValueError("Fix me")
        return self.__class__(self[self.city == "Hometown"])

    def near_(self, distance_in_meters, other_places=None):
        places = get_type_from_registry("places") if other_places is None else other_places
        nearbies = []
        for _, row in places[["lat", "lon"]].drop_duplicates().iterrows():
            nearbies.append(haversine(self.lat, self.lon, *row) < distance_in_meters)
        return self.__class__(self[np.any(nearbies, axis=0)])

    def at(self, time):
        mp = parse_date_tz(time)
        return self[self.index.overlaps(pd.Interval(mp.start_date, mp.end_date))]

    @nlp("end", "how much")
    def sum(self):
        return self.distance.sum()

    @nlp("end", "how long", "how much time")
    def length(self):
        return (self.index.right - self.index.left).to_pytimedelta().sum()

    @nlp("filter", "travel")
    def travel(self):
        return self[self.transporting]

    @nlp(
        "filter",
        "drive",
        "by car",
        "travel by car",
        "travel using the car",
        "by car",
        "going by car",
        "on a car",
        "driving",
    )
    def travel_by_car(self):
        return self.col_contains("Driving", "category")

    @nlp(
        "filter",
        "drive by bus",
        "by bus",
        "travel by bus",
        "using the bus",
        "going by bus",
        "on a bus",
    )
    def travel_by_bus(self):
        return self.col_contains("On a bus", "category")

    @nlp(
        "filter",
        "drive by train",
        "by train",
        "travel by train",
        "using the train",
        "going by train",
        "on a train",
    )
    def travel_by_train(self):
        return self.col_contains("On a train", "category")

    @nlp("filter", "work days", "work-days", "on working days")
    def work_days(self):
        return self[self.time.dt.weekday < 5]

    @nlp("filter", "during work", "during work hours")
    def work_hours(self):
        return self[self._office_hours]

    @nlp("end", "address of", "what is the address", "how to find", "how can i find")
    def what_address(self):
        return self["address"]



"""
Can be scheduled per 2 minutes with:

sysdm create "python -m nostalgia.sources.whereami.scheduler" --extensions "" --timer "*:0/2"
"""

import time
import whereami
import os

if __name__ == "__main__":
    with open(os.path.expanduser("~/nostalgia_data/input/whereami/history.tsv"), "a") as f:
        res = whereami.predict()
        print(res)
        f.write("{}\t{}\n".format(time.time(), res))



import re
from collections import Counter
import just
import requests
from nearnlp.nearnlp import is_noun, is_verb, singularize
import functools
from tok import Tokenizer
from nltk.corpus import stopwords
from nostalgia.enrichers.google.custom_search import google_custom_search

ENGLISH_STOP = set(stopwords.words("english"))

t = Tokenizer(True)
t.drop("<b>", "remove html")
t.drop("<b/>", "remove html")

# can also use qoogle

interesting_keys = set()
for prefix in ["og:", "twitter:", ""]:
    for key in [
        "title",
        "description",
        "name",
        'manufacturer_name',
        'category_name_singular',
        'long_description',
        'snippet',
    ]:
        interesting_keys.add(prefix + key)


def recurse_json(json_data, results=None):
    if results is None:
        results = []
    if isinstance(json_data, dict):
        for k, v in json_data.items():
            if k in interesting_keys and isinstance(v, str) and v:
                results.append(v)
            elif isinstance(v, (dict, list)):
                recurse_json(v, results)
    elif isinstance(json_data, list):
        for x in json_data:
            recurse_json(x, results)
    return results


@functools.lru_cache(2000)
def get_keywords_for_product(product_string):
    # json = {"searchString": product_string}
    # res = requests.get("https://icecat.biz/search/rest/get-products-list", json=json).json()
    res = google_custom_search(product_string)
    zzz = recurse_json(res)
    c = Counter()
    for x in zzz:
        c.update(
            set(
                [
                    singularize(y.lower())
                    for y in t.word_tokenize(x)
                    if re.search("^[a-z]+$", y.lower())
                ]
            )
        )
    # print([(k,v) for k,v in c.most_common(1000) if is_noun(k) and is_verb(k)])
    return [
        x
        for i, x in enumerate(c.most_common(1000))
        if x[0] not in ENGLISH_STOP and len(x[0]) > 2 and x[1] > 1 and is_noun(x[0])
        # if a verb would occur often then probably it is still important
        and (not is_verb(x[0]) or i < 5)
    ]



import os
import just
import string
from datetime import datetime
from dateutil.relativedelta import relativedelta
from dateutil.parser import parse
from datetime import timedelta
import numpy as np
import pandas as pd
from collections import Counter
import re

from nostalgia.times import try_date, tz
from nostalgia.ndf import NDF
from nostalgia.nlp import nlp

digits = set(string.digits)


def old_date(col):
    return pd.to_datetime(
        col.str.slice(0, 14), format="%d-%m-%y %H:%M", errors="coerce"
    ).dt.tz_localize(tz)


class Payments(NDF):
    keywords = ["pay", "buy", "purchase", "spend", "cost"]
    nlp_columns = ["mededelingen", "naam"]
    selected_columns = ["time", "naam", "bedrag", "mededelingen"]

    @classmethod
    def handle_dataframe_per_file(cls, data, fname=None):
        ing = data
        ing["Bedrag (EUR)"] = [float(x.replace(",", ".")) for x in ing["Bedrag (EUR)"]]
        ing["Datum"] = [pd.to_datetime(x, format="%Y%m%d") for x in ing["Datum"]]
        ing["Naam / Omschrijving"] = [x.replace("'", "''") for x in ing["Naam / Omschrijving"]]
        ing["Naam / Omschrijving"] = [x.replace('"', '""') for x in ing["Naam / Omschrijving"]]
        ing["timestamp"] = [
            x if pd.notnull(x) else try_date(y, min_level=3)  # only minutes and below
            for x, y in zip(old_date(ing["Naam / Omschrijving"]), ing.Mededelingen)
        ]
        ing["year"] = [x.year for x in ing.timestamp]

        ing["pas"] = [
            x if y == "Betaalautomaat" and x.startswith("0") else "---"
            for x, y in zip(
                ing.Mededelingen.str.replace(":", "").str.slice(10, 13), ing.MutatieSoort
            )
        ]
        # ing = ing.drop(["Code", "Tegenrekening", "Rekening", "Mededelingen"], axis=1)
        ing.columns = ["bedrag" if x == "Bedrag (EUR)" else x for x in ing.columns]
        ing.columns = ["naam" if x == "Naam / Omschrijving" else x for x in ing.columns]
        ing.columns = [x.lower().replace(" ", "_") for x in ing.columns]

        ing = ing[ing.timestamp.notnull()]
        return ing

    @classmethod
    def load(cls, file_glob="~/Downloads/NL*20*20*.csv", nrows=None, from_cache=True):
        data = cls.latest_file_is_historic(file_glob, nrows=nrows, from_cache=from_cache)
        return cls(data)

    @property
    def expenses(self):
        return self.query("af_bij == 'Af'")

    @nlp("filter", "by me", "i", "my")
    def by_me(self):
        return self.query("pas == '010'")

    @property
    def by_card(self):
        return self.query("pas != '---'")

    @nlp("end", "how much")
    def sum(self):
        return self.bedrag.sum().round(2)

    def equal_to(self, amount):
        return self.query("bedrag == {}".format(amount))

    def amount_between(self, lower_bound, upper_bound):
        return self.query("{} < bedrag < {}".format(lower_bound, upper_bound))



import json
from datetime import datetime
import just
import pandas as pd

from nostalgia.times import tz

from auto_extract import parse_article

from nostalgia.cache import get_cache
from nostalgia.ndf import NDF

from nostalgia.utils import normalize_name

CACHE = get_cache("linked_person")


def getter(dc, key, default=None):
    res = dc.get(key, default)
    if isinstance(res, list):
        res = res[0]
    elif isinstance(res, dict):
        res = json.dumps(res)
    return res


def get_linked_data_jd(art):
    data = None
    try:
        jdata = art.jsonld
    except json.JSONDecodeError:
        return None
    for y in jdata:
        if not y:
            continue
        if isinstance(y, list):
            y = y[0]
        if y.get("@type") != "Person":
            continue
        return {
            'description': getter(y, "description"),
            'startDate': getter(y, "startDate"),
            'endDate': getter(y, "endDate"),
            'location': getter(y, "location"),
            "source": "jsonld",
        }


def get_linked_data_md(art):
    data = None
    for y in art.microdata:
        props = y.get("properties")
        if props is None:
            continue
        tp = str(y.get("type", ""))
        if not tp.endswith("/Person"):
            continue
        return {
            "startDate": getter(props, "startDate"),
            "endDate": getter(props, "endDate"),
            "description": getter(props, "description"),
            "name": "".join(getter(props, "name", "")),
            "location": getter(props, "location"),
            "source": "md",
        }


def get_linked_data(x):
    path = x["path"]
    if path in CACHE:
        return CACHE[path]
    try:
        html = just.read(path)
    except EOFError:
        CACHE[path] = None
        return None
    if not html.strip():
        CACHE[path] = None
        return None
    art = parse_article(html, x["url"])
    linked_data = get_linked_data_md(art)
    if linked_data is None:
        linked_data = get_linked_data_jd(art)
    CACHE[path] = linked_data
    return linked_data


class Person(NDF):
    @classmethod
    def object_to_row(cls, obj):
        row = get_linked_data(obj)
        if row is not None:
            row["time"] = datetime.fromtimestamp(float(obj["time"]), tz=tz)
            row["url"] = obj["url"]
            row["path"] = obj["path"]
            row["keywords"] = ""
            return row

    @classmethod
    def load(cls, file_path="~/nostalgia_data/meta.jsonl", nrows=None, **kwargs):
        person = cls.load_object_per_newline(file_path, nrows)
        return cls(person)


if __name__ == "__main__":
    person = Person.load(nrows=5)

    # https://schema.org/Blog
    # "http://schema.org/blogPost"
    # http://schema.org/WebPage # http://schema.org/BlogPosting

    # exlclude http://schema.org/QAPage

    it = 0
    imo = 0
    nice = 0
    wrong = 0
    from collections import Counter
    import just

    c = Counter()

    for x in just.iread("/home/pascal/nostal_tmp/person.jsonl"):
        if "/Person" in (str(x.get("microdata"))):
            it += 1
            y = x
            score = 0
            mc_count = 0
            for mc in y["microdata"]:
                if mc.get("type") in [
                    'http://schema.org/ImageObject',
                    'http://schema.org/QAPage',
                    'http://schema.org/Movie',
                    'http://schema.org/videoObject',
                    'http://schema.org/Organization',
                    'http://schema.org/VideoObject',
                    'http://schema.org/Question',
                    'http://schema.org/CreativeWork',
                    'http://schema.org/Code',
                ]:
                    continue
                mc_count += 1
                for opt in [
                    'mc["properties"]["author"]',
                    'mc["properties"]["author"]["properties"]["name"]',
                    'mc["properties"]["author"]["value"]',
                    'mc["properties"]["author"][0]["value"]',
                    'mc["properties"]["author"][0]["properties"]["name"]',
                    'mc["properties"]["author"]["properties"]["author"]["properties"]["name"][0]',
                    'mc["properties"]["creator"]["properties"]["name"]',
                    'mc["properties"]["author"][0]["value"]',
                    'mc["properties"]["mainEntity"]["properties"]["author"]["properties"]["name"]',
                    'mc["properties"]["blogPost"]["properties"]["author"]["properties"]["name"]',
                ]:
                    try:
                        x = eval(opt).strip()
                        if not x or x.startswith("http") or "\n" in x:
                            continue
                        print(x)
                        score += 1
                        c[mc["type"].split("/")[-1]] += 1
                        break
                    except KeyboardInterrupt:
                        "a" + 1
                    except:
                        pass
            if mc_count and score == 0:
                if len(y["microdata"]) == 1 and list(y["microdata"][0]) == ["value"]:
                    continue
                wrong += 1
                continue
            if score:
                nice += 1



import os
import just
import zipfile


def load_from_download(ingest_glob, vendor, recent_only=True, delete_existing=True):
    ingest_files = just.glob(ingest_glob)
    if not ingest_files:
        raise ValueError(f"Nothing to extract using {ingest_glob} - Aborting")
    nostalgia_input = "~/nostalgia_data/input/{}".format(vendor)
    if delete_existing:
        just.remove(nostalgia_input, allow_recursive=True)
    elif just.exists(nostalgia_input):
        raise ValueError(f"Cannot overwrite path {nostalgia_input}, pass delete_existing=True")
    fnames = sorted(ingest_files, key=os.path.getctime)
    if recent_only:
        fnames = fnames[-1:]
    for fname in fnames:
        with zipfile.ZipFile(fname, 'r') as zip_ref:
            out = os.path.expanduser(nostalgia_input)
            print("unpacking from", fname, "to", out)
            zip_ref.extractall(out)



import just
import pandas as pd
from nostalgia.ndf import NDF
from nostalgia.times import parse_date_tz
from nostalgia.nlp import nlp


class MijnChipkaart(NDF):
    @property
    def title(self):
        return [x + " - " + y for x, y in zip(self.Vertrek, self.Bestemming)]

    @nlp("end", "how much", "cost")
    def sum(self):
        return self.Bedrag.sum()

    @classmethod
    def load(cls, nrows=None):
        files = just.glob("~/nostalgia_data/input/mijn_chipkaart/*.csv")
        data = pd.concat([pd.read_csv(x, sep=";", nrows=nrows) for x in files])
        data["Bedrag"] = [float(x.replace(",", ".")) for x in data["Bedrag"]]
        data["Datum"] = [
            parse_date_tz(x + " " + y).start_date for x, y in zip(data.Datum, data["Check-uit"])
        ]
        return cls(data)



from nostalgia.ndf import NDF


class Google(NDF):
    vendor = "google"
    ingest_settings = {
        "ingest_glob": "~/Downloads/takeout-20*-*.zip",
        "recent_only": False,
        "delete_existing": False,
    }



import os
import just
from googleapiclient.discovery import build
from googleapiclient.http import HttpError
from datetime import datetime
import diskcache
import dotenv
from nostalgia.cache import get_cache

CACHE = get_cache("google_custom_search")

dotenv.load_dotenv("google_custom_search/.env")
dotenv.load_dotenv(".env")

errored_count = 0


def google_custom_search(search_term, **kwargs):
    global errored_count
    search_term = search_term.lower()
    if search_term in CACHE:
        return CACHE[search_term]
    if errored_count > 4:
        return []
    service = build("customsearch", "v1", developerKey=os.environ.get("MY_API_KEY"))
    try:
        res = service.cse().list(q=search_term, cx=os.environ.get("MY_CSE_ID"), **kwargs).execute()
    except HttpError as e:
        print("error", e)
        errored_count += 1
        return []
    items = res.get("items", [])
    CACHE[search_term] = items
    return items



from nostalgia.ndf import NDF


class Facebook(NDF):
    vendor = "facebook"
    ingest_settings = {
        "ingest_glob": "~/Downloads/facebook-*.zip",
        "recent_only": False,
        "delete_existing": True,
    }



import just

import pandas as pd
import lxml.html

from nostalgia.cache import get_cache
from nostalgia.ndf import NDF

from datetime import datetime
from nostalgia.times import tz


CACHE = get_cache("linked_google_search")


def get_linked_data(x):
    path = x["path"]
    if path in CACHE:
        return CACHE[path]
    try:
        html = just.read(path)
    except EOFError:
        CACHE[path] = None
        return None
    if not html.strip():
        CACHE[path] = None
        return None
    tree = lxml.html.fromstring(html)
    res = tree.xpath("//input[@name='q' and @type='text']")
    if not res:
        linked_data = None
    else:
        linked_data = {"title": res[0].value}
    CACHE[path] = linked_data
    return linked_data


class GoogleSearch(NDF):
    vendor = "web"
    keywords = ["search"]
    nlp_columns = ["title"]
    selected_columns = ["time", "title", "url"]

    @classmethod
    def object_to_row(cls, obj):
        if "google" not in str(obj["url"]) or "/search" not in str(obj["url"]):
            return None
        row = get_linked_data(obj)
        if row is not None:
            row["time"] = datetime.fromtimestamp(float(obj["time"]), tz=tz)
            row["url"] = obj["url"]
            row["path"] = obj["path"]
            row["keywords"] = ""
            return row

    @classmethod
    def load(cls, file_path="~/nostalgia_data/meta.jsonl", nrows=None):
        data = cls.load_object_per_newline(file_path, nrows)
        # google pages found more often?
        data = data[data.title != data.title.shift(1)]
        return cls(data)



from nostalgia.ndf import NDF


class Samsung(NDF):
    vendor = "samsung"
    ingest_settings = {
        "ingest_glob": "~/Downloads/samsunghealth_*.zip",
        "recent_only": True,
        "delete_existing": True,
    }



import numpy as np
from nostalgia.times import datetime_from_format
from datetime import timedelta
from nostalgia.sources.google import Google


def custom_parse(x):
    if not isinstance(x, str):
        return x
    try:
        return datetime_from_format(x, "%Y-%m-%dT%H:%M:%SZ", in_utc=True)
    except:
        return datetime_from_format(x, "%Y-%m-%dT%H:%M:%S.%fZ", in_utc=True)


class PlayMusic(Google):
    @classmethod
    def handle_dataframe_per_file(cls, data, file_path):
        data["_start"] = [custom_parse(x) for x in data["time"]]

        data = data.rename(columns={"description": "artist"})
        data = data[~data.title.str.startswith("Searched for ")]
        data.loc[:, "skip"] = data.title.str.startswith("Skipped ")
        data.loc[:, "listen"] = data.title.str.startswith("Listened to ")
        data = data.sort_values("_start")
        data["skipping_this"] = data.shift(-1).skip
        data["_end"] = data.shift(-1)._start
        data["fake"] = data.skipping_this & (data._end - data._start < timedelta(seconds=15))
        data["long"] = data._end - data._start > timedelta(minutes=10)
        data.loc[data.long, "_end"] = data[data.long]["_start"] + timedelta(minutes=10)
        data = data[~data.fake & data.listen]
        # stripping "Listened to "
        data["title"] = data.title.str.slice(12, None)
        data["duration"] = data._end - data._start
        # clean up
        for d in [
            "products",
            "subtitles",
            "header",
            "titleUrl",
            "skipping_this",
            "skip",
            "listen",
            "fake",
        ]:
            del data[d]
        for name, group in data.groupby(["artist", "title"]):
            if not group.long.any() or group.long.all():
                continue
            max_duration = group.duration.max()
            group.loc[group.long, "duration"] = max_duration
            group.loc[group.long, "_end"] = group._start + max_duration
            group.loc[:, "long"] = False
        # duration cannot be written to parquet, so do it in post process
        del data["duration"]
        return data

    @classmethod
    def load(cls, nrows=None, from_cache=True, **kwargs):
        file_path = (
            "~/nostalgia_data/input/google/Takeout/My Activity/Google Play Music/My Activity.json"
        )
        data = cls.load_data_file_modified_time(file_path, nrows=nrows, from_cache=from_cache)
        data["duration"] = data["_end"] - data["_start"]
        return cls(data)



import os
import pandas as pd
import just
from metadate import parse_date
from datetime import datetime
from nostalgia.ndf import NDF
from nostalgia.times import datetime_from_format


class FitbitSleep(NDF):
    vendor = "fitbit"

    @classmethod
    def load(cls, nrows=None):
        file_glob = "~/nostalgia_data/input/fitbit/*/sleep/*.json"
        objects = []
        for d in just.multi_read(file_glob).values():
            if not d:
                continue
            for x in d:
                data = pd.DataFrame(
                    x["levels"]["data"]
                    + [{'dateTime': x['endTime'], 'level': None, 'seconds': None}]
                )
                data["dateTime"] = [
                    datetime_from_format(x, "%Y-%m-%dT%H:%M:%S.%f") for x in data.dateTime
                ]
                start = data.dateTime.iloc[:-1]
                end = data.dateTime.iloc[1:]
                interval_index = pd.IntervalIndex.from_arrays(start, end)
                data = pd.DataFrame(data.iloc[:-1])
                data = data.set_index(interval_index)
                data["start"] = data.index.left
                data["end"] = data.index.right
                objects.append(data)
                if nrows is not None and data.shape[0] > nrows:
                    break

        data = pd.concat(objects).drop("dateTime", axis=1)

        return cls(data)

    @property
    def asleep(self):
        return self.query("level != 'wake'")



import re
import just
import os
import pandas as pd


def slugify(name):
    return re.sub(r'[\W_]+', '-', name)


def file_modified_since_last(fname, name):
    path = just.make_path("~/nostalgia_data/seen/" + slugify(name) + ".json")
    last_run_mt = float(just.read(path, no_exist=0))
    modified_time = os.path.getmtime(fname)
    if last_run_mt != modified_time:
        return modified_time
    else:
        return None


def get_last_mod_time(name):
    path = "~/nostalgia_data/seen/" + slugify(name) + ".json"
    return just.read(path, no_exist=0.0)


def save_last_mod_time(last_mod_time, name):
    path = "~/nostalgia_data/seen/" + slugify(name) + ".json"
    return just.write(last_mod_time, path)


def get_last_latest_file(name):
    path = "~/nostalgia_data/seen/" + slugify(name) + ".json"
    return just.read(path, no_exist=0)


def save_last_latest_file(latest_file, name):
    path = "~/nostalgia_data/seen/" + slugify(name) + ".json"
    return just.write(latest_file, path)


def get_newline_count(name):
    """ counts by row numbers in a file """
    path = "~/nostalgia_data/seen/" + slugify(name) + ".json"
    return just.read(path, no_exist=0)


def save_newline_count(n, name):
    """ counts by row numbers in a file """
    path = "~/nostalgia_data/seen/" + slugify(name) + ".json"
    return just.write(n, path)


def get_processed_files(name):
    path = "~/nostalgia_data/seen/" + slugify(name) + ".json"
    return set(just.read(path, no_exist=[]))


def save_processed_files(fnames, name):
    path = "~/nostalgia_data/seen/" + slugify(name) + ".json"
    just.write(fnames, path)


def check_seen(name, value):
    path = "~/nostalgia_data/seen/" + slugify(name) + ".json"
    is_new = True
    res = just.read(path, no_exist=False)
    if res:
        if isinstance(value, tuple):
            value = list(value)
        is_new = res != value
    if is_new:
        just.write(value, path)
    return is_new


def make_path(name):
    dir_name = os.path.expanduser("~/nostalgia_data/dfs/")
    just.mkdir(dir_name, 0x700)
    return dir_name + slugify(name)


def save_df(df, name):
    path = make_path(name) + ".parquet"
    df = pd.DataFrame(df)
    for x in ["time", "start", "end"]:
        if x in df:
            df = df.sort_values(x)
            break
    df = df.reset_index(drop=True)
    if "index" in df.columns:
        df.drop("index", axis=1, inplace=True)
    if "level_0" in df.columns:
        df.drop("level_0", axis=1, inplace=True)
    try:
        df.to_parquet(path, compression="zstd", allow_truncated_timestamps=True)
    except Exception as e:
        print("ERROR with", name)
        for x in df.columns:
            print(x, sum([str(y)[:1] == "{" for y in df[x]]))
        raise e
    return df


def load_df(name, nrows=None):
    path = make_path(name) + ".parquet"
    data = pd.read_parquet(path)
    if nrows is not None:
        data = data.iloc[-nrows:]
    return data



import just
from nostalgia.times import datetime_from_format
from nostalgia.sources.samsung import Samsung


class SamsungSleep(Samsung):
    @classmethod
    def handle_dataframe_per_file(cls, data, fname):
        data = data[["start_time", "end_time", "stage"]]
        encoding = {40001: "awake", 40002: "light", 40003: "deep", 40004: "rem"}
        data["stage"] = data.stage.replace(encoding)
        return data

    @classmethod
    def load(cls, nrows=None):
        path = "~/nostalgia_data/input/samsung/samsunghealth_*/com.samsung.health.sleep_stage.*.csv"
        fname = just.glob(path)[0]
        data = cls.load_data_file_modified_time(fname, nrows=nrows, skiprows=1)
        data["start_time"] = [
            datetime_from_format(x, "%Y-%m-%d %H:%M:%S.%f") for x in data["start_time"]
        ]
        data["end_time"] = [
            datetime_from_format(x, "%Y-%m-%d %H:%M:%S.%f") for x in data["end_time"]
        ]
        return cls(data)

    @property
    def asleep(self):
        return self.query("stage != 'awake'")



from nostalgia.ndf import NDF
from nostalgia.nlp import nlp


class ChatInterface(NDF):
    __name__ = ""
    me = ""
    sender_column = ""
    _sender_updated = False

    @nlp("filter", "by me", "i", "my")
    def by_me(self):
        return self[self.sender == "me"]

    @nlp("filter", "by the other", "by someone else")
    def by_other(self):
        return self[self.sender != "me"]

    @property
    def sender(self):
        if not self._sender_updated:
            col = self[self.sender_column]
            self.loc[col == self.me, self.sender_column] = "me"
            self._sender_updated = True
        return self[self.sender_column]



import os
import pandas as pd
from datetime import datetime
from nostalgia.ndf import NDF
from nostalgia.times import tz


def get_time(x):
    y, z = x.split()
    try:
        d = datetime.fromtimestamp(float(y), tz)
    except Exception as e:
        print(e)
        return
    return d, z


class Whereami(NDF):
    nlp_columns = ["name"]

    @classmethod
    def load(cls, file_path="~/nostalgia_data/input/whereami/history.tsv", nrows=None, **kwargs):
        file_path = os.path.expanduser(file_path)
        ndata = []
        nrows = nrows or float("inf")
        it = 0
        with open(file_path) as f:
            for x in f.read().split("\n")[1:]:
                if not x.strip():
                    continue
                x = get_time(x)
                if x is None:
                    continue
                if it > nrows:
                    break
                ndata.append(x)
                it += 1

        whereami = pd.DataFrame(ndata, columns=["start", "name"]).sort_values("start")
        whereami = whereami[whereami.name != "unknown"]
        whereami["end"] = whereami.start.shift(-1)
        whereami["in_range"] = (whereami.end - whereami.start) < pd.Timedelta(minutes=20)
        whereami["new"] = (whereami.name != whereami.name.shift(1)).cumsum()

        groups = []
        group_id = None
        start = None
        name = None
        for s, n, e, ir, new in zip(
            whereami.start, whereami.name, whereami.end, whereami.in_range, whereami.new
        ):
            if group_id is None:
                group_id = new
                start = s
                name = n
            elif group_id != new or not ir:
                groups.append({"start": start, "end": s, "name": name})
                group_id = new
                start = s
                name = n

        return cls(pd.DataFrame(groups))



from selenium import webdriver
import time
from datetime import datetime
from dateutil.relativedelta import relativedelta


def expand_shadow_element(base, tags):
    for tag in tags:
        print(tag)
        element = base.find_element_by_css_selector(tag)
        base = driver.execute_script('return arguments[0].shadowRoot', element)
    return base


if __name__ == "__main__":
    chrome_options = webdriver.ChromeOptions()
    chrome_options.binary_location = "/usr/bin/google-chrome-stable"
    driver = webdriver.Chrome(executable_path="/usr/bin/chromedriver", options=chrome_options)

    url = "https://mijn.ing.nl/banking/service"

    driver.get(url)

    while driver.find_elements_by_xpath("/html/body/ing-app-authentication"):
        time.sleep(1)

    time.sleep(5)

    tags = [
        "dba-app",
        "ing-app-daily-banking-service",
        "ing-orange-service",
        "ing-orange-service-self-control",
    ]
    ele = expand_shadow_element(driver, tags)
    ele.find_element_by_css_selector(
        "div:nth-child(2) > .card__content > ul > li:nth-child(2) > a"
    ).click()

    time.sleep(5)

    tags = [
        "dba-download-transactions-dialog",
        "ing-orange-transaction-download-dialog",
        "ing-uic-dialog-next > section > ing-orange-transaction-download-filter",
        "ing-uic-form > form > div > div > ing-uic-date-input",
        "#viewInput",
    ]
    form = expand_shadow_element(driver, tags)
    inp = form.find_element_by_css_selector(
        "ing-uic-input-container > ing-uic-native-input > input"
    )
    inp.clear()
    inp.send_keys("01-01-2010")

    tags = [
        "dba-download-transactions-dialog",
        "ing-orange-transaction-download-dialog",
        "ing-uic-dialog-next > section > ing-orange-transaction-download-filter",
        "ing-uic-form > form > div > div > ing-uic-date-input[name='endDate']",
        "#viewInput",
    ]
    form = expand_shadow_element(driver, tags)
    inp = form.find_element_by_css_selector(
        "ing-uic-input-container > ing-uic-native-input > input"
    )
    now = datetime.now() - relativedelta(days=1)
    inp.clear()
    inp.send_keys("{}-{}-{}".format(now.day, now.month, now.year))

    tags = [
        "dba-download-transactions-dialog",
        "ing-orange-transaction-download-dialog",
        "ing-uic-dialog-next > section > ing-orange-transaction-download-filter",
    ]
    form = expand_shadow_element(driver, tags)

    form.find_element_by_css_selector(
        "ing-uic-form > form > div > ing-uic-select.format-container"
    ).click()

    time.sleep(2)

    form.find_element_by_css_selector(
        "ing-uic-form > form > div > ing-uic-select.format-container > paper-listbox > ing-uic-item[value='CSV']"
    ).click()

    time.sleep(2)

    form.find_element_by_css_selector(
        "ing-uic-form > form > paper-button[data-type='submit']"
    ).click()

    time.sleep(15)

    driver.close()



import inspect
from collections import defaultdict
from functools import wraps
from nostalgia.times import parse_date_tz

nlp_registry = defaultdict(set)
regex_registry = defaultdict(set)
ts = None

# from nearnlp import Near
# n = Near()
n = None


class ResultInfo(object):
    __slots__ = ('cls', 'role', 'cls_fn', 'col_name', 'orig_word')

    def __init__(self, cls, role, cls_fn=None, col_name=None, orig_word=None):
        self.cls = cls
        self.role = role
        self.cls_fn = cls_fn
        self.col_name = col_name
        self.orig_word = orig_word

    def repr_wrapper(self, x):
        v = None
        if x == "cls_fn":
            try:
                v = str(self.cls_fn).split()[1].split(".")[-1]
            except IndexError:
                pass
        if v is None:
            v = getattr(self, x)
        return v

    @property
    def tp(self):
        if isinstance(self.cls, str):
            return self.cls.split(".")[0]
        return self.cls.__name__

    def __repr__(self):
        try:
            c = getattr(self, "cls").__name__
        except AttributeError:
            c = getattr(self, "cls")
        return "!{}({})".format(
            c,
            ", ".join(
                ["{}={!r}".format(x, self.repr_wrapper(x)) for x in self.__slots__ if x != "cls"]
            ),
        )

    def __hash__(self):
        return hash((self.cls, self.col_name, self.orig_word))

    def __eq__(self, o):
        return (self.cls, self.col_name, self.orig_word) == (o.cls, o.col_name, o.orig_word)


COLUMN_BLACKLIST = set(["did", "was", "the", "a", "are", "i", "from", "hard", "for"])


def att_getter(fn):
    def att_inner(x, *args):
        try:
            return getattr(x, fn.__name__)(*args)
        except TypeError:
            return getattr(x, fn.__name__)()

    return att_inner


# has to be completed still
def regex(role, *keywords):
    def real_decorator(fn):
        for keyword in keywords:
            if isinstance(keyword, str):
                nlp_registry[keyword].add(
                    ResultInfo(fn.__qualname__, role, att_getter(fn), orig_word=keyword)
                )
            else:
                for kw in keyword:
                    nlp_registry[kw].add(
                        ResultInfo(fn.__qualname__, role, att_getter(fn), orig_word=kw)
                    )

        @wraps(fn)
        def wrapper(*args, **kwargs):
            return fn(*args, **kwargs)

        return wrapper

    return real_decorator


def nlp(role, *keywords):
    def real_decorator(fn):
        # this does not work!!!!
        # is_prop = inspect.isdatadescriptor(fn)
        # defin = ", ".join([x for x in inspect.getfullargspec(fn).args if x != "self"])
        # if not is_prop:
        #     defin = "({})".format(defin)
        for keyword in keywords:
            if isinstance(keyword, str):
                nlp_registry[keyword].add(
                    ResultInfo(fn.__qualname__, role, att_getter(fn), orig_word=keyword)
                )
            else:
                for kw in keyword:
                    nlp_registry[kw].add(
                        ResultInfo(fn.__qualname__, role, att_getter(fn), orig_word=kw)
                    )

        @wraps(fn)
        def wrapper(*args, **kwargs):
            return fn(*args, **kwargs)

        return wrapper

    return real_decorator


from textsearch import TextSearch


def get_ts():
    ts = TextSearch("insensitive", "object")
    ts.add(nlp_registry)
    return ts


def at_time_wrapper(mp):
    def at_time(x):
        return x.at_time(mp)

    return at_time


def find_entities(sentence):
    global ts
    if ts is None:
        ts = get_ts()
    mp = parse_date_tz(sentence)
    try:
        ents = ts.findall(sentence)
    except AttributeError:
        raise AttributeError("No entities have been registered")
    # remove metaperiod tokens from otherwise matching
    if mp is not None:
        wrongs = set()
        for l, e in mp.spans:
            wrongs.update(range(l, e))
        ents = [x for x in ents if x.start not in wrongs and x.end not in wrongs]
        ents.append(ResultInfo("MP", "filter", at_time_wrapper(mp), orig_word=" ".join(mp.matches)))
    return ents



from setuptools import find_packages
from setuptools import setup

with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

MAJOR_VERSION = "0"
MINOR_VERSION = "1"
MICRO_VERSION = "27"
VERSION = "{}.{}.{}".format(MAJOR_VERSION, MINOR_VERSION, MICRO_VERSION)

setup(
    name='nostalgia',
    version=VERSION,
    description="nostalgia enables to self-track and gain insights into your life",
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url='https://github.com/nostalgia-dev/nostalgia',
    author='Pascal van Kooten',
    author_email='kootenpv@gmail.com',
    license='MIT',
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Customer Service',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: Microsoft',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Unix',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development',
        'Topic :: Software Development :: Build Tools',
        'Topic :: Software Development :: Debuggers',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Software Distribution',
        'Topic :: System :: Systems Administration',
        'Topic :: Utilities',
    ],
    install_requires=[
        "just",
        "pandas",
        "metadate",
        "tok",
        "ijson",
        "cython",
        "pyarrow",
        "tldextract",
        # "lxml",  # web_history
        # "diskcache",  # web_history
        # "auto_extract",  # web_history
        # "ujson",  # web_history
        # "google-api-python-client",  # offers
        # "python-dotenv",  # offers
        # "natura",  # offers
    ],
    zip_safe=False,
    platforms='any',
)



import re
import os
import pandas as pd
import just
from datetime import datetime
from nostalgia.ndf import NDF
from nostalgia.times import datetime_from_format


def get_day(fname):
    return " ".join(re.sub(".partial_[0-9]+.[0-9]+", "", os.path.basename(fname)).split(".")[1:-1])


class FitbitHeartrate(NDF):
    vendor = "fitbit"

    @classmethod
    def handle_dataframe_per_file(cls, df, fname):
        if df.empty:
            return None
        day = get_day(fname)
        df["time"] = [datetime_from_format(day + " " + x, "%Y %m %d %H:%M:%S") for x in df.time]
        return df

    @classmethod
    def download(cls):
        from nostalgia_fitbit.__main__ import main

        main()

    @classmethod
    def load(cls, nrows=None, **kwargs):
        file_glob = "~/nostalgia_data/input/fitbit/*/heartrate_intraday/**/*.json"
        heartrate = cls.load_dataframe_per_json_file(file_glob, nrows=nrows)
        start = heartrate.time.iloc[:-1]
        end = heartrate.time.iloc[1:]
        interval_index = pd.IntervalIndex.from_arrays(start, end)
        heartrate = pd.DataFrame(heartrate.iloc[:-1])
        heartrate = heartrate.set_index(interval_index)
        heartrate["end"] = heartrate.index.right - pd.Timedelta(seconds=1)
        heartrate = heartrate[heartrate.time != heartrate.end]
        heartrate = heartrate[heartrate.time <= heartrate.end]
        # if nrows is not None:
        #     heartrate = heartrate.iloc[:nrows]
        return cls(heartrate)



import os
import re
import pandas as pd
import numpy as np
from nostalgia.times import now, yesterday, last_week, last_month, last_year, parse_date_tz
from metadate import is_mp
from datetime import timedelta
import just
from nostalgia.nlp import nlp_registry, nlp, n, COLUMN_BLACKLIST, ResultInfo
from nostalgia.utils import get_token_set, normalize_name, view

from nostalgia.extracter import load_from_download
from nostalgia.file_caching import save_df, load_df
from nostalgia.times import datetime_from_timestamp
from nostalgia.cache import get_cache
from nostalgia.data_loading import Loader


def ab_overlap_cd(a, b, c, d):
    return ((a < d) & (b > c)) | ((b > c) & (d > a))


def ab_overlap_c(a, b, c):
    return (a <= c) & (c <= b)


def join_time_naive(locs, df, **window_kwargs):
    tmp = []
    if not locs.inferred_time:
        locs.infer_time()
    for _, row in locs.iterrows():
        if df.start is not None:
            if locs._start_col is None:
                start = row[locs._time_col] - pd.Timedelta(**window_kwargs)
                end = row[locs._time_col] + pd.Timedelta(**window_kwargs)
            else:
                start = row[locs._start_col]
                end = row[locs._end_col]
            res = df[ab_overlap_cd(df.start, df.end, start, end)]
        else:
            if locs._start_col is None:
                start = row[locs._time_col] - pd.Timedelta(**window_kwargs)
                end = row[locs._time_col] + pd.Timedelta(**window_kwargs)
            else:
                start, end = row.start, row.end
            res = df[df.time.between(start, end)]
        if not res.empty:
            # import pdb
            # pdb.set_trace()
            tmp.append(res)
    if not tmp:
        return df[:0]
    tmp = pd.concat([pd.DataFrame(x) for x in tmp])
    try:
        tmp = tmp.drop_duplicates()
    except TypeError:
        tmp = tmp[~tmp.astype(str).duplicated()]
    return tmp


def join_time_index(locs, df, **window_kwargs):
    try:
        return df[df.time.apply(lambda x: locs.index.contains(x))]
    except TypeError:
        print("warning")
        return join_time_naive(locs, df, **window_kwargs)


def join_time(locs, df, **window_kwargs):
    if df.start is not None:
        return join_time_naive(locs, df, **window_kwargs)
    if not str(df.index.dtype).startswith("interval"):
        return join_time_naive(locs, df, **window_kwargs)
    W = locs.shape[0]
    L = df.shape[0]
    y = 0.001 + W * -0.318 + L * 0.013
    if y > 300:
        p = 1
    else:
        exponent = np.exp(y)
        p = exponent / (1 + exponent)
    if p > 0.5:
        fn = join_time_naive
    else:
        fn = join_time_index
    return fn(locs, df, **window_kwargs)


registry = {}


def get_type_from_registry(tp):
    for key, value in registry.items():
        if key.endswith(tp):
            return value


def time(x):
    return x.time


def col_contains_wrapper(word, col):
    def col_contains(x):
        return x.col_contains(word, col)

    return col_contains


class NDF(Loader, pd.DataFrame):
    keywords = []
    nlp_columns = []
    nlp_when = True
    selected_columns = []
    vendor = None

    def __init__(self, data):
        super().__init__(data)
        C = self.__class__
        self.num_times = None
        self._start_col, self._time_col, self._end_col = None, None, None
        self.inferred_time = False
        # IF BIG BREAK, THEN THIS IS HERE
        # if self.df_name not in registry:
        #     registry[self.df_name] = self
        if self.df_name != "results":
            if self.df_name not in registry:
                registry[self.df_name] = self
            # if the new data is smaller than what is in the registry, do not overwrite the registry
            elif data.shape[0] > registry[self.df_name].shape[0]:
                registry[self.df_name] = self
        if n is None:
            return
        seen_keyword_keywords = set()
        for kw in self.keywords:
            for w in n.simple_extend(kw):
                if w and isinstance(w, str):
                    nlp_registry[w].add(ResultInfo(C, "start", orig_word=kw))
                    seen_keyword_keywords.add(w)
            nlp_registry[kw].add(ResultInfo(C, "start", orig_word=kw))
            seen_keyword_keywords.add(kw)
        # nlp_keywords_column = []
        # if "keywords" in data:
        #     nlp_keywords_column = ["keywords"]
        BLACKLIST = seen_keyword_keywords.union(COLUMN_BLACKLIST)
        for col in self.nlp_columns:
            words = get_token_set(self[col].unique())
            for word in words:
                if word in BLACKLIST:
                    continue
                for w in n.simple_extend(word):
                    if w in BLACKLIST:
                        continue
                    if w and isinstance(w, str):
                        nlp_registry[w].add(
                            ResultInfo(C, "filter", col_contains_wrapper(word, col), col, word)
                        )
                if n.is_verb(word):
                    for w in n.get_verbs(word).values():
                        if w in BLACKLIST:
                            continue
                        if w in words or not isinstance(w, str):
                            continue
                        nlp_registry[w].add(
                            ResultInfo(C, "filter", col_contains_wrapper(word, col), col, word)
                        )
                if n.is_verb(word):
                    for w in n.get_verbs(word).values():
                        if w in BLACKLIST:
                            continue
                        if w in words or not isinstance(w, str):
                            continue
                        nlp_registry[w].add(
                            ResultInfo(C, "filter", col_contains_wrapper(word, col), col, word)
                        )
                nlp_registry[word].add(
                    ResultInfo(C, "filter", col_contains_wrapper(word, col), col, word)
                )
        if self.nlp_when:
            nlp_registry["when"].add(ResultInfo(C, "end", time, orig_word="when"))

    @classmethod
    def ingest(cls):
        load_from_download(vendor=cls.vendor, **cls.ingest_settings)

    @classmethod
    def get_normalized_name(cls):
        return normalize_name(cls.__name__)

    @classmethod
    def load_df(cls, nrows):
        return load_df(cls.get_normalized_name(), nrows)

    @classmethod
    def save_df(cls, df, name=None):
        return save_df(df, name or cls.get_normalized_name())

    @classmethod
    def load(cls, nrows=None):
        df = cls.load_df(nrows)
        return cls(df)

    @property
    def df_name(self):
        name = normalize_name(self.__class__.__name__)
        if self.vendor is not None and not name.startswith(self.vendor):
            name = self.vendor + "_" + name
        return name

    @classmethod
    def df_label(cls):
        return normalize_name(cls.__name__).replace("_", " ").title()

    @property
    def time(self):
        if not self.inferred_time:
            self.infer_time()
            self.inferred_time = True
        if self._time_col == "time":
            return self._get_item_cache(self._time_col)
        if self._time_col not in self:
            return None
        return getattr(self, self._time_col)

    # @property
    # def _constructor(self):
    #     return self.__class__

    def duration_longer_than(self, **timedelta_kwargs):
        return self[(self.end - self.time) >= timedelta(**timedelta_kwargs)]

    def duration_shorter_than(self, **timedelta_kwargs):
        return self[(self.end - self.time) <= timedelta(**timedelta_kwargs)]

    def take_from(self, registry_ending, col_name):
        for registry_type in registry:
            if not registry_type.endswith(registry_ending):
                continue
            # TODO: loop over columns, so we only do index lookup once
            # TODO: do not only try self.time but also self.end
            new_name = registry_ending + "_" + col_name
            if new_name in self.columns:
                return self[new_name]
            tp = get_type_from_registry(registry_type)
            results = []
            if not self.inferred_time:
                self.infer_time()
            for x in self[self._time_col]:
                try:
                    res = tp.loc[x]
                    if not isinstance(res, pd.Series):
                        res = res.iloc[0]
                    res = res[col_name]
                except (KeyError, TypeError):
                    res = np.nan
                results.append(res)
            self[new_name] = results
            return self[new_name]

    def add_heartrate(self):
        return self.take_from("heartrate", "value")

    def heartrate_range(self, low, high=None):
        if "heartrate_value" not in self.columns:
            self.add_heartrate()
        if high is not None and low is not None:
            return self[(self["heartrate_value"] >= low) & self["heartrate_value"] < high]
        if low is not None:
            return self[self["heartrate_value"] >= low]
        if high is not None:
            return self[self["heartrate_value"] < high]

    def heartrate_above(self, value):
        return self.heartrate_range(value)

    def heartrate_below(self, value):
        return self.heartrate_range(None, value)

    @classmethod
    def get_schema(cls, *args, **kwargs):
        sample = cls.load(*args, nrows=5, **kwargs)
        return {k: v for k, v in zip(sample.columns, sample.dtypes)}

    @property
    def start(self):
        if not self.inferred_time:
            self.infer_time()
            self.inferred_time = True
        if self._start_col == "start":
            return self._get_item_cache(self._start_col)
        if self._start_col not in self:
            return None
        return getattr(self, self._start_col)

    @property
    def _office_days(self):
        return self.time.dt.weekday < 5

    @property
    def _office_hours(self):
        if (self.time.dt.hour == 0).all():
            raise ValueError("Hours are not set, thus unreliable. Use `office_days` instead?")
        if self.start is not None:
            return np.array(
                [
                    (x >= 8 and x <= 17 and y >= 8 and y <= 17)
                    for x, y in zip(self.start.dt.hour, self.end.dt.hour)
                ]
            )
        return np.array([(x >= 8 and x <= 17) for x in self.time.dt.hour])

    @property
    def in_office_days(self):
        return self[self._office_days]

    @property
    def in_office_hours(self):
        return self[self._office_hours]

    @property
    def during_office_hours(self):
        return self[self._office_hours & self._office_days]

    @property
    def outside_office_hours(self):
        return self[~(self._office_hours & self._office_days)]

    @property
    def end(self):
        if not self.inferred_time:
            self.infer_time()
            self.inferred_time = True
        if self._end_col == "end":
            return self._get_item_cache(self._end_col)
        if self._end_col not in self:
            return None
        return getattr(self, self._end_col)

    def time_level(self, col):
        if (col.dt.microsecond != 0).any():
            return 4
        if (col.dt.second != 0).any():
            return 3
        if (col.dt.minute != 0).any():
            return 2
        if (col.dt.hour != 0).any():
            return 1
        return 0

    def infer_time(self):
        if self.__class__.__name__ == "Results":
            self._start_col, self._time_col, self._end_col = "start", "start", "end"
            return
        times = [x for x, y in zip(self.columns, self.dtypes) if "datetime" in str(y)]
        levels = [self.time_level(self[x]) for x in times]
        if not levels:
            raise ValueError(f"No datetime found in {self.__class__.__name__}")
        max_level = max(levels)
        # workaround
        # start: 10:00:00
        # end:   10:00:59
        times = [t for t, l in zip(times, levels) if l == max_level or (l == 2 and max_level == 3)]
        num_times = len(times)
        self.num_times = num_times
        if num_times == 0:
            self._start_col, self._time_col, self._end_col = None, None, None
        elif num_times == 1:
            self._start_col, self._time_col, self._end_col = None, times[0], None
        elif num_times == 2:
            col1, col2 = times
            sub = self[self[col1].notnull() & self[col2].notnull()]
            a, b = sub[col1], sub[col2]
            if (a > b).all():
                col1, col2 = col2, col1
            elif not (a <= b).all():
                raise ValueError(
                    "Not strictly one col higher than other with dates, can't determine"
                )
            if col1 == "end" and col2 == "start":
                col2, col1 = col1, col2
            self._start_col, self._time_col, self._end_col = col1, col1, col2
        else:
            msg = 'infer time failed: there can only be 1 or 2 datetime columns at the same granularity.'

            raise Exception(msg + " Found: " + str(times))

    def when_at(self, other, **window_kwargs):
        if isinstance(other, str):
            other = get_type_from_registry("places").containing(other)
        return self.__class__(join_time(other, self, **window_kwargs))

    def browsing(self, other, **window_kwargs):
        if isinstance(other, str):
            other = get_type_from_registry("browser").containing(other)
        return self.__class__(join_time(other, self, **window_kwargs))

    def at_night(self, start=22, end=8):
        return self.between_hours(start, end)

    def between_hours(self, start=22, end=8):
        if self._start_col is not None:
            return self[(self.start.dt.hour > start) | (self.end.dt.hour < end)]
        return self[(self.time.dt.hour > start) & (self.time.dt.hour < end)]

    @property
    def when_asleep(self):
        return self.__class__(join_time(registry["sleep"].asleep, self))

        # browser["session"] = browser.time.diff().apply(lambda x: x.seconds > 20 * 60).cumsum()+10
        # session, start, end = zip(*[(name, group.iloc[0].time - pd.Timedelta(minutes=10), group.iloc[-1].time + pd.Timedelta(minutes=10)) for name, group in browser.groupby("session")])
        # df = pd.DataFrame({"session": session}, index=pd.IntervalIndex.from_arrays(start, end))
        # df.index.get_loc(pd.Timestamp(parse_date_tz("2019-09-18 18:40:56.729633").start_date))

    def col_contains(self, string, col_name, case=False, regex=False, na=False):
        return self[self[col_name].str.contains(string, case=case, regex=regex, na=na)]

    def __getitem__(self, key):
        new = super().__getitem__(key)
        if isinstance(new, pd.DataFrame):
            new = self.__class__(new)
        return new

    @property
    def text_cols(self):
        return [x for x, t in zip(self.columns, self.dtypes) if t == np.dtype('O')]

    @nlp("filter", "by me", "i", "my")
    def by_me(self):
        return self

    def containing(self, string, col_name=None, case=False, regex=True, na=False, bound=True):
        """
        Filters using string in all string columns when col_name is None, otherwise in just that one
        When `bound=True` it means to add word boundaries to the regex.
        case=True is whether to be case-sensitive
        regex=True means to treat string as regex
        na=False means to consider NaN to be considered False
        """
        if regex and bound:
            string = r"\b" + string + r"\b"
        if col_name is not None:
            return self.col_contains(string, col_name, case, regex, na)
        bool_cols = [
            self[x].str.contains(string, case=case, regex=regex, na=na) for x in self.text_cols
        ]
        bool_array = bool_cols[0]
        for b in bool_cols[1:]:
            bool_array = np.logical_or(bool_array, b)
        return self.__class__(self[bool_array])

    def query(self, expr):
        return self.__class__(super().query(expr))

    def as_simple(self, max_n=None):
        data = {
            "title": self.df_name,  # default, to be overwritten
            "url": None,
            "start": None,
            "end": None,
            # "body": None,
            "type": self.df_name,
            "interval": True,
            "sender": None,
            "value": getattr(self, "value", None),
            "index_loc": self.index,
        }
        for x in ["title", "name", "naam", "subject", "url", "content", "text", "value"]:
            res = getattr(self, x, None)
            if res is not None:
                data["title"] = res
                break
        res = getattr(self, "sender", None)
        if res is not None:
            data["sender"] = res
        for x in ["url", "path", "file"]:
            res = getattr(self, x, None)
            if res is not None:
                data["url"] = res
                break
        for x in ["start", "time", "timestamp"]:
            res = getattr(self, x, None)
            if res is not None:
                data["start"] = res
                break
        for x in ["end"]:
            res = getattr(self, x, None)
            if res is not None:
                data["end"] = res - pd.Timedelta(microseconds=1)
                break
        if data["end"] is None:
            data["end"] = data["start"] + pd.Timedelta(minutes=5)
            data["interval"] = False
        try:
            data = pd.DataFrame(data).sort_values("start")
            if max_n is not None:
                data = data.iloc[-max_n:]
            return data
        except ValueError:
            raise ValueError(f"No fields are mapped for {self.__class__.__name__}")

    def _select_at_day(self, day_or_class):
        if isinstance(day_or_class, pd.DataFrame):
            days = day_or_class.time.dt.date.unique()
            return self.time.dt.date.isin(days)
        elif isinstance(day_or_class, (list, tuple, set, pd.Series)):
            return self.time.dt.date.isin(set(day_or_class))
        else:
            mp = parse_date_tz(day_or_class)
            return (self.time.dt.date >= mp.start_date.date()) & (
                self.time.dt.date < mp.end_date.date()
            )

    def at_day(self, day_or_class):
        return self[self._select_at_day(day_or_class)]

    def not_at_day(self, day_or_class):
        return self[~self._select_at_day(day_or_class)]

    @property
    def last_day(self):
        return self[(self.time > yesterday()) & (self.time < now())]

    @property
    def yesterday(self):
        return self[(self.time > yesterday()) & (self.time < now())]

    @property
    def last_week(self):
        return self[(self.time > last_week()) & (self.time < now())]

    @property
    def last_month(self):
        return self[(self.time > last_month()) & (self.time < now())]

    @property
    def last_year(self):
        return self[(self.time > last_year()) & (self.time < now())]

    def near(self, s):
        if isinstance(s, NDF) and s.df_name.endswith("places"):
            selection = s
        else:
            selection = get_type_from_registry("places").containing(s)
        return self.when_at(selection)

    def to_place(self):
        results = []
        places = get_type_from_registry("places")
        for time in self.time:
            try:
                results.append(places.iloc[places.index.get_loc(time)].iloc[0])
            except (TypeError, KeyError):
                pass
        return places.__class__(results)

    def in_a(self, s):
        return self.near(s)

    def read(self, index):
        return just.read(self.path[index])

    def view(self, index):
        view(self.path[index])

    def head(self, *args, **kwargs):
        return self.__class__(super().head(*args, **kwargs))

    def tail(self, *args, **kwargs):
        return self.__class__(super().tail(*args, **kwargs))

    @nlp("end", "how many", "how many times", "how often")
    def count(self):
        return self.shape[0]

    @property
    def at_home(self):
        self.take_from("places", "category")
        return self[self["places_category"] == "Home"]

    @property
    def at_work(self):
        self.take_from("places", "category")
        return self[self["places_category"] == "Work"]

    def at_time(self, start, end=None, sort_diff=True, **window_kwargs):
        if is_mp(start):
            start = start.start_date
            end = start.end_date
        elif isinstance(start, str) and end is None:
            mp = parse_date_tz(start)
            start = mp.start_date
            end = mp.end_date
        elif isinstance(start, str) and isinstance(end, str):
            mp = parse_date_tz(start)
            start = mp.start_date
            mp = parse_date_tz(end)
            end = mp.end_date
        elif end is None and window_kwargs:
            end = start
        elif end is None:
            raise ValueError(
                "Either a metaperiod, a date string, 2 times, or time + window_kwargs."
            )
        self.infer_time()
        if window_kwargs:
            start = start - pd.Timedelta(**window_kwargs)
            end = end + pd.Timedelta(**window_kwargs)
        if self._start_col is None:
            res = self[ab_overlap_c(start, end, self[self._time_col])]
        else:
            res = self[ab_overlap_cd(self[self._start_col], self[self._end_col], start, end)]
        if not res.empty and sort_diff:
            # avg_time = start + (end - start) / 2
            # res["sort_score"] = -abs(res[self._time_col] - avg_time)
            # res = res.sort_values('sort_score').drop('sort_score', axis=1)
            res["sort_score"] = res[self._time_col]
            res = res.sort_values('sort_score').drop('sort_score', axis=1)
        return self.__class__(res)

    when = when_at

    @property
    def duration(self):
        return self.end - self.start

    def sort_values(
        self, by, axis=0, ascending=True, inplace=False, kind='quicksort', na_position='last'
    ):
        return self.__class__(
            pd.DataFrame.sort_values(self, by, axis, ascending, inplace, kind, na_position)
        )

    @nlp("filter", "last", "last time", "most recently")
    def last(self):
        _ = self.time  # to get inferred time if not set
        col = self._time_col or self._start_col
        return self.__class__(self.sort_values(col, na_position="last", ascending=False).iloc[:1])

    # maybe which/what could be showing only unique?
    @nlp("end", "show", "show me", "show me the", "show the", "what")
    def show_me(self):
        _ = self.time  # to get inferred time if not set
        col = self._time_col or self._start_col
        return self.__class__(self.sort_values(col, na_position="last", ascending=False))

    def at(self, time_or_place):
        if isinstance(time_or_place, NDF) and time_or_place.df_name.endswith("places"):
            return self.when_at(time_or_place)
        if isinstance(time_or_place, str):
            mp = parse_date_tz(time_or_place)
            if mp:
                start = mp.start_date
                end = mp.end_date
                return self.at_time(start, end)
            else:
                return self.when_at(get_type_from_registry("places").containing(time_or_place))
        raise ValueError("neither time nor place was passed")

    def to_html(self):
        if self.selected_columns:
            data = pd.DataFrame({x: getattr(self, x) for x in self.selected_columns})
            return data.to_html()
        return super().to_html()

    def get_type_from_registry(self, tp):
        for key, value in rergistry.items():
            if key.endswith(tp):
                return value

    def __call_(self):
        return self


# ov = ab_overlap_cd(
#     places.start,
#     places.end,
#     parse("2015-02-05 21:53:09.581001+00:00"),
#     parse("2015-02-05 21:53:13.881000+00:00"),
# )


class Results(NDF):
    @classmethod
    def merge(cls, *dfs, max_n=None):
        data = pd.concat([x.as_simple(max_n) for x in dfs])
        data = data.set_index("start")
        data = data.sort_values("start", na_position="first")
        data["start"] = data.index
        return Results(data)

    def get_original(self, index_loc):
        row = self.iloc[index_loc]
        rec = registry[row.type].loc[row["index_loc"]]
        # rec.add_heartrate() # not here, but at the results level.. but then just to display
        # if a certain value
        if not isinstance(rec, pd.Series):
            rec = rec.iloc[0]
        return rec



import os
import just
from datetime import datetime
import pandas as pd
from nostalgia.times import tz
from nostalgia.utils import format_latlng
from nostalgia.sources.google import Google


class Photos(Google):
    @classmethod
    def load(cls, nrows=None):
        photo_glob = "~/nostalgia_data/input/google/Takeout/Google Photos/*/*"
        pics = []
        nrows = nrows or float("inf")
        rows = 0
        for fname in just.glob(photo_glob):
            if fname.endswith(".json"):
                continue
            try:
                meta = just.read(fname + ".json")
            except FileNotFoundError:
                continue
            if rows == nrows:
                break
            date = datetime.fromtimestamp(int(meta['photoTakenTime']['timestamp']), tz)
            lat, lon = format_latlng(
                (meta['geoData']['latitude'], meta['geoData']['longitude'])
            ).split(", ")
            title = meta["title"]
            pics.append(
                {"path": "file://" + fname, "lat": lat, "lon": lon, "title": title, "time": date}
            )
            rows += 1

        pics = pd.DataFrame(pics)
        return cls(pics)



# import getpass
# from nostalgia.selenium import get_driver
# import lxml.html

# driver = get_driver(
#     credentials={"username": "kootenpv@gmail.com", "password": getpass.getpass()},
#     login_url="https://www.ah.nl/mijn/inloggen",
# )

# url = "https://www.ah.nl/producten/eerder-gekocht?sorting=last_bought"

# driver.get(url)

# tree = lxml.html.fromstring(driver.page_source)

# prices = [x.text_content() for x in tree.xpath("//div[@class='product-price']")]
# descriptions = [
#     x.text_content().replace("\xad", "")
#     for x in tree.xpath("//div[@class='product-description']/h1")
# ]
# links = [
#     "https://www.ah.nl" + x for x in tree.xpath("//article[contains(@class, 'product')]/a/@href")
# ]



import just
import pandas as pd
from nostalgia.interfaces.chat import ChatInterface
from nostalgia.times import datetime_from_format

offset = len("01/10/2019, 20:02 - ")


class WhatsappChat(ChatInterface):
    """# WhatsApp Chat
    WhatsApp allows you to export a single conversation (.txt) by email using a mobile app. Below official instructions for each platform:

    - [Android](https://faq.whatsapp.com/en/android/23756533/)
    - iPhone (does not work)
    - [Windows Phone](https://faq.whatsapp.com/en/wp/22548236)

   ### Other method

    datas = []
    for fname in just.glob("~/Downloads/*.csv"):
        try:
            with open(fname) as f:
                chars = f.read(30)
                if "Date1;Date2;Time;UserPhone" not in chars:
                    continue
                datas.append(pd.read_csv(fname, sep=";"))
        except PermissionError:
            continue

   ### Create instance
   Click below on the `+` sign and fill in the path to the WhatsApp chat file."""

    vendor = "whatsapp"
    me = ""
    sender_column = "sender"

    @classmethod
    def load(cls, nrows=None, **kwargs):
        old_text = ""
        results = []
        nrows = nrows or float("inf")
        for file_path in just.glob("~/nostalgia_data/input/whatsapp/*.txt"):
            row = 0
            for line in just.iread(file_path):
                try:
                    time = datetime_from_format(line[:offset], "%d/%m/%Y, %H:%M - ")
                except ValueError:
                    old_text += line + "\n"
                    continue
                line = old_text + line[offset:]
                old_text = ""
                try:
                    if line.startswith("Messages to this chat and calls are now secured"):
                        continue
                    sender, text = line.split(": ", 1)
                except ValueError:
                    print("ERR", line)
                    continue
                if line:
                    if row > nrows:
                        break
                    row += 1
                    results.append((time, sender, text))

        df = pd.DataFrame(results, columns=["time", "sender", "text"])
        # hack "order" into minute data
        same_minute = df.time == df.shift(1).time
        seconds = []
        second_prop = 0
        for x in same_minute:
            if x:
                second_prop += 1
            else:
                second_prop = 0
            seconds.append(pd.Timedelta(seconds=60 * second_prop / (second_prop + 1)))
        df["time"] = df["time"] + pd.Series(seconds)
        return cls(df)
